'''
fit_tscan_gui:
Graphical User Interface for fit_tscan utility

:copyright: 2021-2022 by pistack (Junho Lee).
:license: LGPL3.
'''

from sys import platform
import re
import tkinter as tk
import tkinter.messagebox as msg
import tkinter.filedialog as fd
import tkinter.scrolledtext as sc
import numpy as np
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.backends.backend_tkagg import NavigationToolbar2Tk
from matplotlib.backend_bases import key_press_handler
from matplotlib.figure import Figure
from ..driver import fit_transient_exp, fit_transient_dmp_osc, fit_transient_both
from ..driver import save_TransientResult

FITDRIVER = {'decay': fit_transient_exp, 'osc': fit_transient_dmp_osc,
'both': fit_transient_both}

float_sep_comma = re.compile('([\+\-]?[0-9]+[.]?[0-9]*[,]\s*)*[\+\-]?[0-9]+[.]?[0-9]*\s*')
isfloat = re.compile('[\+\-]?[0-9]+[.]?[0-9]*\s*')

# check font
if platform == "linux" or platform == "linux2":
    widgetfont = 'Liberation Sans'
else:
    widgetfont = 'Arial'


class PlotDataWidget:
    '''
    Class for widget which plot data
    '''
    # These codes are based on Matplotlib example "Embedding in Tk"

    def __init__(self, master):

        self.top = tk.Toplevel(master.root)
        self.top.title('Plot Data')

        self.fig = Figure(figsize=(8, 4), dpi=100)

        # immutable
        self.ax = self.fig.add_subplot()
        self.ax.set_xlabel('Time Delay')
        self.ax.set_ylabel('Intensity')
        self.ax.grid(True)

        # mutable
        self.ax.set_title(f'{master.fname[0]}')
        self.ln, = self.ax.plot(master.t[0], master.intensity[0][:, 0], mfc='none',
        color='black', marker='o')
        self.poly = self.ax.fill_between(master.t[0],
        master.intensity[0][:, 0]-master.eps[0][:, 0],
        master.intensity[0][:, 0]+master.eps[0][:, 0], alpha=0.5, color='black')

        # set canvas
        self.canvas = FigureCanvasTkAgg(self.fig, master=self.top)
        self.canvas.draw()

        self.toolbar = \
            NavigationToolbar2Tk(self.canvas, self.top, pack_toolbar=False)
        self.toolbar.update()

        self.canvas.mpl_connect('key_press_event', key_press_handler)

        self.slider_update = tk.Scale(self.top, from_=1, to_=len(master.fname),
        orient=tk.HORIZONTAL, command=lambda val: self.update_plot(master, int(val)),
        label='File index')

        self.slider_update.pack(side=tk.BOTTOM)
        self.toolbar.pack(side=tk.BOTTOM, fill=tk.X)
        self.canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=True)
        self.top.mainloop()

    # --- Update plot when one moves slider

    def update_plot(self, master, val):
        self.ax.set_title(f'{master.fname[val-1]}')
        self.ln.set_data(master.t[val-1], master.intensity[val-1][:, 0])
        self.poly.remove()
        self.poly = self.ax.fill_between(master.t[val-1],
        master.intensity[val-1][:, 0]-master.eps[val-1][:, 0],
        master.intensity[val-1][:, 0]+master.eps[val-1][:, 0], alpha=0.5,
        color='black')
        xmin = np.min(master.t[val-1])
        xmax = np.max(master.t[val-1])
        ymin = np.min(master.intensity[val-1][:, 0])
        ymax = np.max(master.intensity[val-1][:, 0])
        errmax = np.max(master.eps[val-1][:, 0])
        self.ax.set_xlim((xmin, xmax))
        self.ax.set_ylim((ymin-5*errmax, ymax+5*errmax))
        self.canvas.draw()


class PlotFitWidget:
    '''
    Class for wideget which plots fitting result
    '''
    # These codes are based on Matplotlib example "Embedding in Tk"

    def __init__(self, master):

        self.top = tk.Toplevel(master.root)
        self.top.title('Plot Fitting Result')

        self.fig = Figure(figsize=(8, 8), dpi=100)

        # immutable
        # fit
        self.ax_fit = self.fig.add_subplot(211)
        self.ax_fit.set_xlabel('Time Delay')
        self.ax_fit.set_ylabel('Intensity')
        self.ax_fit.grid(True)

        # residual
        self.ax_res = self.fig.add_subplot(212)
        self.ax_res.set_xlabel('Time Delay')
        self.ax_res.set_ylabel('Residual')
        self.ax_res.grid(True)

        # mutable

        # fit
        self.ax_fit.set_title(f'{master.fname[0]}')
        self.ln_fit, = self.ax_fit.plot(master.t[0], master.result['fit'][0][:, 0],
        color='red')
        self.ln_data, = self.ax_fit.plot(master.t[0], master.intensity[0][:, 0], mfc='none',
        color='black', marker='o')
        self.poly_data = self.ax_fit.fill_between(master.t[0],
        master.intensity[0][:, 0]-master.eps[0][:, 0],
        master.intensity[0][:, 0]+master.eps[0][:, 0], alpha=0.5, color='black')

        # residual
        if master.fit_mode_var.get() in ['decay', 'osc']:
            self.ln_res, = self.ax_res.plot(master.t[0],
            master.intensity[0][:, 0]-master.result['fit'][0][:, 0],
            mfc='none', color='black', marker='o')
            self.poly_res = self.ax_res.fill_between(master.t[0],
            master.intensity[0][:, 0]-master.result['fit'][0][:, 0]-master.eps[0][:, 0],
            master.intensity[0][:, 0]-master.result['fit'][0][:, 0]+master.eps[0][:, 0],
            alpha=0.5, color='black')
        else:
            self.ln_fit_osc, = self.ax_res.plot(master.t[0], master.result['fit_osc'][0][:, 0],
            color='red')
            self.ln_res, = self.ax_res.plot(master.t[0],
            master.intensity[0][:, 0]-master.result['fit_decay'][0][:, 0],
            mfc='none', color='black', marker='o')
            self.poly_res = self.ax_res.fill_between(master.t[0],
            master.intensity[0][:, 0]-master.result['fit_decay'][0][:, 0]-master.eps[0][:, 0],
            master.intensity[0][:, 0]-master.result['fit_decay'][0][:, 0]+master.eps[0][:, 0],
            alpha=0.5, color='black')



        # set canvas
        self.canvas = FigureCanvasTkAgg(self.fig, master=self.top)
        self.canvas.draw()

        self.toolbar = \
            NavigationToolbar2Tk(self.canvas, self.top, pack_toolbar=False)
        self.toolbar.update()

        self.canvas.mpl_connect('key_press_event', key_press_handler)

        self.slider_update = tk.Scale(self.top, from_=1, to_=len(master.fname),
        orient=tk.HORIZONTAL, command=lambda val: self.update_plot(master, int(val)),
        label='File index')

        self.slider_update.pack(side=tk.BOTTOM)
        self.toolbar.pack(side=tk.BOTTOM, fill=tk.X)
        self.canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=True)
        self.top.mainloop()

    # --- Update plot when one moves slider

    def update_plot(self, master, val):

        # update fitting window
        self.ax_fit.set_title(f'{master.fname[val-1]}')
        self.ln_fit.set_data(master.t[val-1], master.result['fit'][val-1][:, 0])
        self.ln_data.set_data(master.t[val-1], master.intensity[val-1][:, 0])
        self.poly_data.remove()
        self.poly_data = self.ax_fit.fill_between(master.t[val-1],
        master.intensity[val-1][:, 0]-master.eps[val-1][:, 0],
        master.intensity[val-1][:, 0]+master.eps[val-1][:, 0], alpha=0.5,
        color='black')
        xmin = np.min(master.t[val-1])
        xmax = np.max(master.t[val-1])
        ymin_fit = np.min(master.intensity[val-1][:, 0])
        ymax_fit = np.max(master.intensity[val-1][:, 0])
        errmax = np.max(master.eps[val-1][:, 0])
        self.ax_fit.set_xlim((xmin, xmax))
        self.ax_fit.set_ylim((ymin_fit-5*errmax, ymax_fit+5*errmax))

        # update residual window

        if master.fit_mode_var.get() in ['decay', 'osc']:
            self.ln_res.set_data(master.t[val-1],
            master.intensity[val-1][:, 0]-master.result['fit'][val-1][:, 0])
            self.poly_res.remove()
            self.poly_res = self.ax_res.fill_between(master.t[val-1],
            master.intensity[val-1][:, 0]-master.result['fit'][val-1][:, 0]-master.eps[val-1][:, 0],
            master.intensity[val-1][:, 0]-master.result['fit'][val-1][:, 0]+master.eps[val-1][:, 0],
            alpha=0.5, color='black')
            ymax_res = np.max(master.intensity[val-1][:, 0]-master.result['fit'][val-1][:, 0])
            ymin_res = np.min(master.intensity[val-1][:, 0]-master.result['fit'][val-1][:, 0])
        else:
            self.ln_fit_osc.set_data(master.t[val-1], master.result['fit_osc'][val-1][:, 0])
            self.ln_res.set_data(master.t[val-1],
            master.intensity[val-1][:, 0]-master.result['fit_decay'][val-1][:, 0])
            self.poly_res.remove()
            self.poly_res = self.ax_res.fill_between(master.t[0],
            master.intensity[val-1][:, 0]-master.result['fit_decay'][val-1][:, 0]-master.eps[val-1][:, 0],
            master.intensity[val-1][:, 0]-master.result['fit_decay'][val-1][:, 0]+master.eps[val-1][:, 0],
            alpha=0.5, color='black')
            ymax_res = np.max(master.intensity[val-1][:, 0]-master.result['fit_decay'][val-1][:, 0])
            ymin_res = np.min(master.intensity[val-1][:, 0]-master.result['fit_decay'][val-1][:, 0])

        self.ax_res.set_xlim((xmin, xmax))
        self.ax_res.set_ylim((ymin_res-2*errmax, ymax_res+2*errmax))

        self.canvas.draw()

class FitReportWidget:
    '''
    Class for reporting fitting result
    '''
    def __init__(self):
        self.root = tk.Tk()
        self.root.title('Fitting Report')
        self.report_space = sc.ScrolledText(self.root,
        width = 80, height = 40)
        self.report_space.configure(state='disabled')
        self.report_space.grid(column=0)

    def update_result(self, result):
        self.report_space.configure(state='normal')
        self.report_space.delete(1.0, 'end')
        self.report_space.insert(tk.INSERT, str(result))
        self.report_space.configure(state='disabled')

    def mainloop(self):
        self.root.mainloop()

    def quit(self):
        self.root.quit()

class FitTscanGuiWidget:
    '''
    Class for fit_tscan gui wrapper
    '''
    def __init__(self):
        self.root = tk.Tk()
        self.root.title('Fit Tscan Gui')
        self.parameter_window = tk.Tk()
        self.parameter_window.title('Initial fitting parameter')
        self.fit_report_window = FitReportWidget()
        self.result = None

        # -- define necessary variables
        self.irf_var = tk.StringVar()
        self.fit_mode_var = tk.StringVar()
        self.base_var = tk.IntVar()
        self.fix_irf_var = tk.IntVar()
        self.fix_t0_var = tk.IntVar()
        self.custom_bd_var = tk.IntVar()
        self.glb_opt_var = tk.StringVar()
        self.fname = []

        # --- fitting model selection window
        self.fit_mode_label = tk.Label(self.root, text='Select fitting mode',
        padx=90, pady=10, font=(widgetfont, 12))
        self.fit_mode_label.grid(column=0, row=0, columnspan=3)
        self.decay_mode = tk.Checkbutton(self.root, text='decay',
        variable = self.fit_mode_var, onvalue = 'decay', offvalue='')
        self.decay_mode.grid(column=0, row=1)
        self.dmp_osc_mode = tk.Checkbutton(self.root, text='damped osc',
        variable = self.fit_mode_var, onvalue = 'dmp_osc', offvalue='')
        self.dmp_osc_mode.grid(column=1, row=1)
        self.both_mode = tk.Checkbutton(self.root, text='decay+dmp_osc',
        variable = self.fit_mode_var, onvalue = 'both', offvalue='')
        self.both_mode.grid(column=2, row=1)

        # --- irf model selection window
        self.irf_label = tk.Label(self.root, text='Select Type of irf',
        padx=90, pady=10, font=(widgetfont, 12))
        self.irf_label.grid(column=0, row=2, columnspan=3)
        self.irf_g = tk.Checkbutton(self.root, text='gaussian', variable=self.irf_var,
        onvalue='g', offvalue='')
        self.irf_g.grid(column=0, row=3)
        self.irf_c = tk.Checkbutton(self.root, text='cauchy', variable=self.irf_var,
        onvalue='c', offvalue='')
        self.irf_c.grid(column=1, row=3)
        self.irf_pv = tk.Checkbutton(self.root, text='pseudo voigt', variable=self.irf_var,
        onvalue='pv', offvalue='')
        self.irf_pv.grid(column=2, row=3)

        # --- global optimization Algorithm
        self.glb_opt_label = tk.Label(self.root, text='Global optimization Methods',
        padx=60, pady=10, font=(widgetfont, 12))
        self.glb_opt_label.grid(column=0, row=4, columnspan=2)
        self.glb_opt_ampgo = tk.Checkbutton(self.root, text='AMPGO', variable=self.glb_opt_var,
        onvalue='ampgo', offvalue='')
        self.glb_opt_ampgo.grid(column=0, row=5)
        self.glb_opt_basin = tk.Checkbutton(self.root, text='Basinhopping',
        variable=self.glb_opt_var,
        onvalue='basinhopping', offvalue='')
        self.glb_opt_basin.grid(column=1, row=5)

        # --- miscellaneous options
        self.option_label = tk.Label(self.root, text='Miscellaneous Options',
        padx=120, pady=10, font=(widgetfont, 12))
        self.option_label.grid(column=0, row=6, columnspan=4)
        self.include_base_check = tk.Checkbutton(self.root, text='base',
        variable=self.base_var, onvalue=1, offvalue=0)
        self.include_base_check.grid(column=0, row=7)
        self.fix_irf_check = tk.Checkbutton(self.root, text='fix_irf',
        variable=self.fix_irf_var, onvalue=1, offvalue=0)
        self.fix_irf_check.grid(column=1, row=7)
        self.fix_t0_check = tk.Checkbutton(self.root, text='fix_t0',
        variable=self.fix_t0_var, onvalue=1, offvalue=0)
        self.fix_t0_check.grid(column=2, row=7)
        self.custom_bd_check = tk.Checkbutton(self.root, text='custom bound',
        variable=self.custom_bd_var, onvalue=1, offvalue=0)
        self.custom_bd_check.grid(column=3, row=7)

        # --- Read file to fit
        self.label_file = tk.Label(self.root, text='Browse Files to fit',
        padx=120, pady=10, font=(widgetfont, 12))
        self.label_file.grid(column=0, row=8, columnspan=4)
        self.print_file_num = tk.Canvas(self.root, width=320, height=20, bd=5,
        bg='white')
        self.print_file_num.grid(column=0, row=9, columnspan=2)
        self.button_file = tk.Button(self.root, width=30, bd=5, text='browse',
        command=self.browse_file)
        self.button_file.grid(column=2, row=9)
        self.button_plot = tk.Button(self.root, width=30, bd=5, text='plot',
        command=self.plot_file)
        self.button_plot.grid(column=3, row=9)

        self.param_button = tk.Button(self.root,
        text='Parameters', command=self.view_param,
        font=(widgetfont, 12), bg='green', padx=30, pady=10, bd=5, fg='white')
        self.param_button.grid(column=0, row=10)

        self.run_button = tk.Button(self.root,
        text='Run', command=self.run_script,
        font=(widgetfont, 12), bg='blue', padx=30, pady=10, bd=5, fg='white')
        self.run_button.grid(column=1, row=10)

        self.save_button = tk.Button(self.root,
        text='Save', command=self.save_script,
        font=(widgetfont, 12), bg='black', padx=30, pady=10, bd=5, fg='white')
        self.save_button.grid(column=2, row=10)

        self.exit_button = tk.Button(self.root,
        text='Exit', command=self.exit_script,
        font=(widgetfont, 12), bg='red', padx=30, pady=10, bd=5, fg='white')
        self.exit_button.grid(column=3, row=10)

        # ---- Parameters Depending on fitting Model ----

        # Parameter
        self.label_paramter = tk.Label(self.parameter_window, text='Paramter',
        font=(widgetfont, 15), padx=120, pady=10)
        self.label_paramter.grid(column=0, row=0, columnspan=4)
        self.label_fwhm_G = tk.Label(self.parameter_window, text='fwhm_G (irf)',
        padx=30, pady=10, font=(widgetfont, 12))
        self.label_fwhm_G.grid(column=0, row=1)
        self.entry_fwhm_G = tk.Entry(self.parameter_window, width=10, bd=1)
        self.entry_fwhm_G.grid(column=0, row=2)

        self.label_fwhm_L = tk.Label(self.parameter_window, text='fwhm_L (irf)',
        padx=30, pady=10, font=(widgetfont, 12))
        self.label_fwhm_L.grid(column=1, row=1)
        self.entry_fwhm_L = tk.Entry(self.parameter_window, width=10, bd=1)
        self.entry_fwhm_L.grid(column=1, row=2)

        self.label_t0 = tk.Label(self.parameter_window,
        text='Insert initial time zero parameter (t01,t02,...)',
        padx=90, pady=10, font=(widgetfont, 12))
        self.label_t0.grid(column=0, row=3, columnspan=3)
        self.entry_t0 = tk.Entry(self.parameter_window, width=90, bd=5)
        self.entry_t0.grid(column=0, row=4, columnspan=3)

        self.label_tau = tk.Label(self.parameter_window,
        text='Insert initial life time parameter (tau1,tau2,...)',
        padx=90, pady=10, font=(widgetfont, 12))
        self.label_tau.grid(column=0, row=5, columnspan=3)
        self.entry_tau = tk.Entry(self.parameter_window, width=90, bd=5)
        self.entry_tau.grid(column=0, row=6, columnspan=3)

        self.label_osc = tk.Label(self.parameter_window,
        text='Insert initial osc period parameter (T_osc1,T_osc2,...)',
        padx=90, pady=10, font=(widgetfont, 12))
        self.label_osc.grid(column=0, row=7, columnspan=3)
        self.entry_osc = tk.Entry(self.parameter_window, width=90, bd=5)
        self.entry_osc.grid(column=0, row=8, columnspan=3)

        self.label_dmp = tk.Label(self.parameter_window,
        text='Insert initial damping lifetime parameter (dmp1,dmp2,...)',
        padx=90, pady=10, font=(widgetfont, 12))
        self.label_dmp.grid(column=0, row=9, columnspan=3)
        self.entry_dmp = tk.Entry(self.parameter_window, width=90, bd=5)
        self.entry_dmp.grid(column=0, row=10, columnspan=3)

        # Parameter Bound

        self.label_bound = tk.Label(self.parameter_window, text='Bounds',
        font=(widgetfont, 15), padx=120, pady=10)
        self.label_bound.grid(column=0, row=11, columnspan=4)

        # fwhm_G (lower)
        self.label_bd_l_fwhm_G = tk.Label(self.parameter_window, text='fwhm_G (lower)',
        padx=30, pady=10, font=(widgetfont, 12))
        self.label_bd_l_fwhm_G.grid(column=0, row=12)
        self.entry_bd_l_fwhm_G = tk.Entry(self.parameter_window, width=10, bd=1)
        self.entry_bd_l_fwhm_G.grid(column=0, row=13)

        # fwhm_G (upper)
        self.label_bd_u_fwhm_G = tk.Label(self.parameter_window, text='fwhm_G (upper)',
        padx=30, pady=10, font=(widgetfont, 12))
        self.label_bd_u_fwhm_G.grid(column=1, row=12)
        self.entry_bd_u_fwhm_G = tk.Entry(self.parameter_window, width=10, bd=1)
        self.entry_bd_u_fwhm_G.grid(column=1, row=13)

        # fwhm_L (lower)
        self.label_bd_l_fwhm_L = tk.Label(self.parameter_window, text='fwhm_L (lower)',
        padx=30, pady=10, font=(widgetfont, 12))
        self.label_bd_l_fwhm_L.grid(column=2, row=14)
        self.entry_bd_l_fwhm_L = tk.Entry(self.parameter_window, width=10, bd=1)
        self.entry_bd_l_fwhm_L.grid(column=2, row=15)

        # fwhm_L (upper)
        self.label_bd_u_fwhm_L = tk.Label(self.parameter_window, text='fwhm_L (upper)',
        padx=30, pady=10, font=(widgetfont, 12))
        self.label_bd_u_fwhm_L.grid(column=2, row=14)
        self.entry_bd_u_fwhm_L = tk.Entry(self.parameter_window, width=10, bd=1)
        self.entry_bd_u_fwhm_L.grid(column=2, row=15)

        # bound time zero
        self.label_bd_l_t0 = tk.Label(self.parameter_window,
        text='Lower bound of time zero (t01,t02,...)',
        padx=60, pady=10, font=(widgetfont, 12))
        self.label_bd_l_t0.grid(column=0, row=16, columnspan=2)
        self.entry_bd_l_t0 = tk.Entry(self.parameter_window, width=60, bd=5)
        self.entry_bd_l_t0.grid(column=0, row=17, columnspan=2)

        self.label_bd_u_t0 = tk.Label(self.parameter_window,
        text='Upper bound of time zero (t01,t02,...)',
        padx=60, pady=10, font=(widgetfont, 12))
        self.label_bd_u_t0.grid(column=2, row=16, columnspan=2)
        self.entry_bd_u_t0 = tk.Entry(self.parameter_window, width=60, bd=5)
        self.entry_bd_u_t0.grid(column=2, row=17, columnspan=2)

        # bound tau
        self.label_bd_l_tau = tk.Label(self.parameter_window,
        text='Lower bound of life time (tau1,tau2,...)',
        padx=60, pady=10, font=(widgetfont, 12))
        self.label_bd_l_tau.grid(column=0, row=18, columnspan=2)
        self.entry_bd_l_tau = tk.Entry(self.parameter_window, width=60, bd=5)
        self.entry_bd_l_tau.grid(column=0, row=19, columnspan=2)

        self.label_bd_u_tau = tk.Label(self.parameter_window,
        text='Upper bound of life time (tau1,tau2,...)',
        padx=60, pady=10, font=(widgetfont, 12))
        self.label_bd_u_tau.grid(column=2, row=18, columnspan=2)
        self.entry_bd_u_tau = tk.Entry(self.parameter_window, width=60, bd=5)
        self.entry_bd_u_tau.grid(column=2, row=19, columnspan=2)

        # bound osc
        self.label_bd_l_osc = tk.Label(self.parameter_window,
        text='Lower bound of osc period (T_osc1,T_osc2,...)',
        padx=60, pady=10, font=(widgetfont, 12))
        self.label_bd_l_osc.grid(column=0, row=20, columnspan=2)
        self.entry_bd_l_osc = tk.Entry(self.parameter_window, width=60, bd=5)
        self.entry_bd_l_osc.grid(column=0, row=21, columnspan=2)

        self.label_bd_u_osc = tk.Label(self.parameter_window,
        text='Upper bound of osc period (T_osc1,T_osc2,...)',
        padx=60, pady=10, font=(widgetfont, 12))
        self.label_bd_u_osc.grid(column=2, row=20, columnspan=2)
        self.entry_bd_u_osc = tk.Entry(self.parameter_window, width=60, bd=5)
        self.entry_bd_u_osc.grid(column=2, row=21, columnspan=2)

        # bound dmp
        self.label_bd_l_dmp = tk.Label(self.parameter_window,
        text='Lower bound of osc life time (dmp1,dmp2,...)',
        padx=60, pady=10, font=(widgetfont, 12))
        self.label_bd_l_dmp.grid(column=0, row=22, columnspan=2)
        self.entry_bd_l_dmp = tk.Entry(self.parameter_window, width=60, bd=5)
        self.entry_bd_l_dmp.grid(column=0, row=23, columnspan=2)

        self.label_bd_u_dmp = tk.Label(self.parameter_window,
        text='Upper bound of osc life time (dmp1,dmp2,...)',
        padx=60, pady=10, font=(widgetfont, 12))
        self.label_bd_u_dmp.grid(column=2, row=22, columnspan=2)
        self.entry_bd_u_dmp = tk.Entry(self.parameter_window, width=60, bd=5)
        self.entry_bd_u_dmp.grid(column=2, row=23, columnspan=2)

        self.hide_init_param_option()

        self.root.mainloop()
        self.parameter_window.mainloop()
        self.fit_report_window.mainloop()

    # --- Browse root directory of fitting file
    def browse_file(self):
        self.file_lst = fd.askopenfilenames(parent=self.root, title='Choose files',
        filetypes=(('any files', '*'), ('text files', '*.txt'), ('data files', '*.dat')))
        self.print_file_num.delete('all')
        self.print_file_num.create_text(200, 15, text=f'{len(self.file_lst)} number of files are loaded')
        self.fname = []
        self.t = []
        self.intensity = []
        self.eps = []
        for fn in self.file_lst:
            tmp = np.genfromtxt(fn)
            self.t.append(tmp[:, 0])
            self.intensity.append(tmp[:, 1].reshape((tmp.shape[0], 1)))
            self.eps.append(tmp[:, 2].reshape((tmp.shape[0], 1)))
            self.fname.append(fn.split('/')[-1])

    def plot_file(self):

        if len(self.fname) == 0:
            msg.showerror('Error', 'Please load files')
        else:
            PlotDataWidget(self)


    # --- hide fitting parameter entry
    def hide_init_param_option(self):
        # bound
        self.label_bound.grid_remove()

        # irf option
        self.label_fwhm_G.grid_remove()
        self.entry_fwhm_G.grid_remove()
        self.label_fwhm_L.grid_remove()
        self.entry_fwhm_L.grid_remove()

        self.label_bd_l_fwhm_G.grid_remove()
        self.label_bd_u_fwhm_G.grid_remove()
        self.entry_bd_l_fwhm_G.grid_remove()
        self.entry_bd_u_fwhm_G.grid_remove()

        self.label_bd_l_fwhm_L.grid_remove()
        self.label_bd_u_fwhm_L.grid_remove()
        self.entry_bd_l_fwhm_L.grid_remove()
        self.entry_bd_u_fwhm_L.grid_remove()

        # t0 option
        self.label_t0.grid_remove()
        self.entry_t0.grid_remove()

        self.label_bd_l_t0.grid_remove()
        self.entry_bd_l_t0.grid_remove()
        self.label_bd_u_t0.grid_remove()
        self.entry_bd_u_t0.grid_remove()

        # tau option
        self.label_tau.grid_remove()
        self.entry_tau.grid_remove()

        self.label_bd_l_tau.grid_remove()
        self.entry_bd_l_tau.grid_remove()
        self.label_bd_u_tau.grid_remove()
        self.entry_bd_u_tau.grid_remove()

        # osc option
        self.label_osc.grid_remove()
        self.entry_osc.grid_remove()

        self.label_bd_l_osc.grid_remove()
        self.entry_bd_l_osc.grid_remove()
        self.label_bd_u_osc.grid_remove()
        self.entry_bd_u_osc.grid_remove()

        # dmp option
        self.label_dmp.grid_remove()
        self.entry_dmp.grid_remove()

        self.label_bd_l_dmp.grid_remove()
        self.entry_bd_l_dmp.grid_remove()
        self.label_bd_u_dmp.grid_remove()
        self.entry_bd_u_dmp.grid_remove()

    # --- prepare to fit
    def view_param(self):

        # hide all
        self.hide_init_param_option()

        if self.custom_bd_var.get():
            self.label_bound.grid()

        if not self.irf_var.get():
            msg.showerror('Error',
            'Please select the type of irf before clicking ready button')
            return

        # show irf option
        if self.irf_var.get() in ['g', 'pv']:
            self.label_fwhm_G.grid()
            self.entry_fwhm_G.grid()

            if self.custom_bd_var.get():
                self.label_bd_l_fwhm_G.grid()
                self.label_bd_u_fwhm_G.grid()
                self.entry_bd_l_fwhm_G.grid()
                self.entry_bd_u_fwhm_G.grid()
        
        if self.irf_var.get() in ['c', 'pv']:
            self.label_fwhm_L.grid()
            self.entry_fwhm_L.grid()

            if self.custom_bd_var.get():
                self.label_bd_l_fwhm_L.grid()
                self.label_bd_u_fwhm_L.grid()
                self.entry_bd_l_fwhm_L.grid()
                self.entry_bd_u_fwhm_L.grid()

        # show t0 option
        self.label_t0.grid()
        self.entry_t0.grid()

        if self.custom_bd_var.get():
            self.label_bd_l_t0.grid()
            self.label_bd_u_t0.grid()
            self.entry_bd_l_t0.grid()
            self.entry_bd_u_t0.grid()

        # show initial life time related option
        if not self.fit_mode_var.get():
            msg.showerror('Error',
            'Please select the fitting model before clicking ready button')
            return

        if self.fit_mode_var.get() in ['decay', 'both']:
            self.label_tau.grid()
            self.entry_tau.grid()

            if self.custom_bd_var.get():
                self.label_bd_l_tau.grid()
                self.label_bd_u_tau.grid()
                self.entry_bd_l_tau.grid()
                self.entry_bd_u_tau.grid()

        if self.fit_mode_var.get() in ['osc', 'both']:

            self.label_osc.grid()
            self.entry_osc.grid()

            self.label_dmp.grid()
            self.entry_dmp.grid()

            if self.custom_bd_var.get():
                self.label_bd_l_osc.grid()
                self.label_bd_u_osc.grid()
                self.entry_bd_l_osc.grid()
                self.entry_bd_u_osc.grid()
                self.label_bd_l_dmp.grid()
                self.label_bd_u_dmp.grid()
                self.entry_bd_l_dmp.grid()
                self.entry_bd_u_dmp.grid()
    
    def handle_float(self, entry, entry_name):
        if entry.get():
            test = entry.get()
            if isfloat.match(test):
                value = float(test)
            else:
                msg.showerror('Error',
                f'{entry_name} should be single float number')
                return False
        else:
            msg.showerror('Error',
            f'Please click Parameter Button and enter {entry_name}')
            return False
        return value

    def handle_float_field(self, entry, entry_name):
        if entry.get():
            tst = entry.get()
            if float_sep_comma.match(tst):
                value = np.array(list(map(float, tst.split(','))))
            else:
                msg.showerror('Error',
                f'{entry_name} should be single float or floats seperated by comma.')
                return False
        else:
            msg.showerror('Error',
            f'Please enter {entry_name}')
            return False
        return value
                
    def handle_irf(self):
        if self.irf_var.get() == 'g':
            fwhm = self.handle_float(self.entry_fwhm_G, 'fwhm_G')
        elif self.irf_var.get() == 'c':
            fwhm = self.handle_float(self.entry_fwhm_L, 'fwhm_L')
        elif self.irf_var.get() == 'pv':
            if self.entry_fwhm_G.get() and self.entry_fwhm_L.get():
                fwhm_G_init = self.entry_fwhm_G.get()
                fwhm_L_init = self.entry_fwhm_L.get()
                if isfloat.match(fwhm_G_init) and isfloat.match(fwhm_L_init):
                    fwhm = [float(fwhm_G_init), float(fwhm_L_init)]
                else:
                    msg.showerror('Error',
                    'Both fwhm_G and fwhm_L field should be single float number')
                    return False
            else:
                msg.showerror('Error',
                'Please click Parameter button and enter both initial fwhm_G and fwhm_L values')
                return False
        else:
            msg.showerror('Error',
            'Please select irf model and click paramter button')
        return fwhm

    def handle_t0(self):
        t0 = self.handle_float_field(self.entry_t0, 't0')
        if not isinstance(t0, np.ndarray):
            return t0
        if t0.size != len(self.file_lst):
            msg.showerror('Error',
            'Number of initial time zero should be same as number of files to fit.')
            return False
        return t0

    def handle_tau(self):
        if self.entry_tau.get():
            str_tau = self.entry_tau.get()
            if float_sep_comma.match(str_tau):
                tau = np.array(list(map(float, str_tau.split(','))))
            else:
                msg.showerror('Error',
                'initial life time constant tau should be single float or floats seperated by comma.')
                return False
        else:
            if self.base_var.get():
                return True
            else:
                msg.showerror('Error',
                'Please enter initial life time constant')
                return False
        return tau

    def handle_osc(self):
        osc = self.handle_float_field(self.entry_osc, 'osc period')
        return osc

    def handle_dmp(self):
        dmp = self.handle_float_field(self.entry_dmp, 'dmp life time')
        return dmp
    
    def handle_bd_irf(self):
        bound_fwhm = None
        if self.irf_var.get() in ['g','pv']:
            bd_l_fwhm_G = self.handle_float(self.entry_bd_l_fwhm_G, 'fwhm_G (lower)')
            if not bd_l_fwhm_G:
                return None
            bd_u_fwhm_G = self.handle_float(self.entry_bd_u_fwhm_G, 'fwhm_G (upper)')
            if not bd_u_fwhm_G:
                return None
            bound_fwhm_G = (bd_l_fwhm_G, bd_u_fwhm_G)
        if self.irf_var.get() in ['c', 'pv']:
            bd_l_fwhm_L = self.handle_float(self.entry_bd_l_fwhm_L, 'fwhm_L (lower)')
            if not bd_l_fwhm_L:
                return None
            bd_u_fwhm_L = self.handle_float(self.entry_bd_u_fwhm_L, 'fwhm_L (upper)')
            if not bd_u_fwhm_L:
                return None
            bound_fwhm_L = (bd_l_fwhm_L, bd_u_fwhm_L)
        
        if self.irf_var.get() == 'g':
            bound_fwhm = [bound_fwhm_G]
        elif self.irf_var.get() == 'c':
            bound_fwhm = [bound_fwhm_L]
        elif self.irf_var.get() == 'pv':
            bound_fwhm = [bound_fwhm_G, bound_fwhm_L]
        return bound_fwhm

    def handle_bd_field(self, param, entry_bd_l, entry_bd_u, entry_name):
        bd_l = self.handle_float_field(entry_bd_l, f'{entry_name} (lower)')
        if not isinstance(bd_l, np.ndarray):
            return None
        bd_u = self.handle_float_field(entry_bd_u, f'{entry_name} (upper)')
        if not isinstance(bd_u, np.ndarray):
            return None
        
        if bd_l.size != bd_u.size:
            msg.showerror('Error', 'The size of lower and upper bound should be same')
            return None
        if bd_l.size != param.size:
            msg.showerror('Error', 'The size of bound should be same as size of initial parameter')
            return None

        return list(zip(bd_l, bd_u))

    def run_script(self):

        # check files are loaded
        if len(self.fname) == 0:
            msg.showerror('Error', 'Please read files before fitting')
            return
        
        dargs = []
        kwargs_key = []
        kwargs_val = []
        kwargs = {}

        bound_fwhm = None
        bound_t0 = None
        bound_tau = None
        bound_dmp = None
        bound_osc = None

        # set initial fwhm
        irf = self.irf_var.get()
        fwhm = self.handle_irf()
        if not fwhm:
            return
        # set bound fwhm
        if self.custom_bd_var.get():
            bound_fwhm = self.handle_bd_irf()
        kwargs_key.append('bound_fwhm')

        # set initial t0
        t0 = self.handle_t0()
        if not isinstance(t0, np.ndarray):
            return
        
        if self.custom_bd_var.get():
            bound_t0 = self.handle_bd_field(t0, self.entry_bd_l_t0,
        self.entry_bd_u_t0, 'time zero')
        kwargs_key.append('bound_t0')

        # handle fix_irf option
        if self.fix_irf_var.get():
            if irf in ['g', 'c']:
                bound_fwhm = [(fwhm, fwhm)]
            elif irf == 'pv':
                bound_fwhm = [(fwhm[0], fwhm[0]),
                (fwhm[1], fwhm[1])]
        
        kwargs_val.append(bound_fwhm)

        # handle fix_t0 option
        if self.fix_t0_var.get():
            bound_t0 = t0.size*[None]

            for i in range(t0.size):
                bound_t0[i] = (t0[i], t0[i])
        
        kwargs_val.append(bound_t0)

        mode = self.fit_mode_var.get()

        if mode in ['decay', 'both']:
            base = self.base_var.get()
            tau = self.handle_tau()
            if isinstance(tau, np.ndarray) or tau:
                if not isinstance(tau, np.ndarray):
                    tau = None
                else:
                    if self.custom_bd_var.get():
                        bound_tau = self.handle_bd_field(tau,
                        self.entry_bd_l_tau, self.entry_bd_u_tau,
                        'tau')
                    kwargs_key.append('bound_tau')
                    kwargs_val.append(bound_tau)
            else:
                return
            dargs.append(tau)

            if mode == 'decay':
                dargs.append(base)

        if mode in ['osc', 'both']:
            dmp = self.handle_dmp()
            if not dmp:
                return
            osc = self.handle_osc()
            if not osc:
                return
            if dmp.size != osc.size:
                msg.showerror('Error', 'The number of damping constant and oscillation period should be same')
                return
            
            if self.custom_bd_var.get():
                bound_dmp = self.handle_bd_field(dmp, self.entry_bd_l_dmp,
                self.entry_bd_u_dmp, 'dmp life time')
                bound_osc = self.handle_bd_field(osc, self.entry_bd_l_osc,
                self.entry_bd_u_osc, 'osc period')
            
            if mode == 'osc':
                kwargs_key.append('bound_tau')
                kwargs_val.append(bound_dmp)
                kwargs_key.append('bound_period')
                kwargs_val.append(bound_osc)
            else:
                kwargs_key.append('bound_tau_osc')
                kwargs_val.append(bound_dmp)
                kwargs_key.append('bound_period_osc')
                kwargs_val.append(bound_osc)

            dargs.append(dmp)
            dargs.append(osc)

        if mode == 'both':
            dargs.append(base)

        if not self.glb_opt_var.get():
            glb_opt = None
        else:
            glb_opt = self.glb_opt_var.get()
        
        for k,v in zip(kwargs_key, kwargs_val):
            kwargs[k] = v

        self.result = FITDRIVER[mode](irf, fwhm, t0, *dargs, method_glb=glb_opt,
        **kwargs,
        name_of_dset=np.array(self.fname), t=self.t, intensity=self.intensity, eps=self.eps)

        self.fit_report_window.update_result(self.result)
        PlotFitWidget(self)

        return
    
    def save_script(self):
        if self.result is None:
            msg.showerror('Error', 'Please click save button after click run button')
            return
        
        save_folder = fd.askdirectory()
        save_h5_name = save_folder + '/' + save_folder.split('/')[-1]
        save_TransientResult(self.result, save_h5_name)
        
        with open(f'{save_folder}/fit_summary.txt', 'w', encoding='utf-8') as f:
            f.write(str(self.result))
            
        for i in range(len(self.result['t'])):
            coeff_fmt = self.result['eps'][i].shape[1]*['%.8e']
            fit_fmt = (1+self.result['eps'][i].shape[1])*['%.8e']
            coeff_header_lst = []
            fit_header_lst = ['time_delay']
            res_save = np.vstack(
                (self.result['t'][i], self.result['res'][i][:, 0], self.result['eps'][i][:, 0])).T
            np.savetxt(f"{save_folder}/res_{self.result['name_of_dset'][i]}.txt", res_save,
                       fmt=['%.8e', '%.8e', '%.8e'],
                       header=f"time_delay \t res_{self.result['name_of_dset'][i]} \t eps")
            fit_header_lst.append(f"fit_{self.result['name_of_dset'][i]}")
            coeff_header_lst.append(f"tscan_{self.result['name_of_dset'][i]}")
            
            fit_header = '\t'.join(fit_header_lst)
            coeff_header = '\t'.join(coeff_header_lst)
            
            np.savetxt(f"{save_folder}/weight_{self.result['name_of_dset'][i]}.txt", 
            self.result['c'][i], fmt=coeff_fmt, header=coeff_header)
            
            if self.result['model'] in ['dmp_osc', 'both']:
                np.savetxt(f"{save_folder}/phase_{self.result['name_of_dset'][i]}.txt", 
                self.result['phase'][i], fmt=coeff_fmt, header=coeff_header)
            
            fit_save = np.vstack((self.result['t'][i], self.result['fit'][i].T)).T
            np.savetxt(f"{save_folder}/fit_{self.result['name_of_dset'][i]}.txt",
            fit_save, fmt=fit_fmt, header=fit_header)
            
            if self.result['model'] == 'both':
                fit_decay_save = np.vstack((self.result['t'][i], self.result['fit_decay'][i].T)).T
                np.savetxt(f"{save_folder}/fit_decay_{self.result['name_of_dset'][i]}.txt",
                fit_decay_save, fmt=fit_fmt, header=fit_header)
                fit_osc_save = np.vstack(
                    (self.result['t'][i], self.result['fit_osc'][i].T)).T
                np.savetxt(f"{save_folder}/fit_osc_{self.result['name_of_dset'][i]}.txt", 
                fit_osc_save, fmt=fit_fmt, header=fit_header)

    def exit_script(self):
        self.parameter_window.quit()
        self.fit_report_window.quit()
        self.root.quit()

def fit_tscan_gui():
    FitTscanGuiWidget()





'''
fit tscan:
submodule to fit tscan data
Using sum of exponential decay convolved with
normalized gaussian distribution
normalized cauchy distribution
normalized pseudo voigt profile
(Uniform fwhm and Mixing parameter eta is calculated according to
Journal of Applied Crystallography. 33 (6): 1311–1316.)

:copyright: 2021-2022 by pistack (Junho Lee).
:license: LGPL3.
'''


from typing import Optional
import os
from pathlib import Path
import argparse
import numpy as np
import matplotlib.pyplot as plt
from ..res import set_bound_tau
from ..driver import TransientResult, save_TransientResult
from ..driver import fit_transient_exp, fit_transient_raise
from ..driver import fit_transient_dmp_osc, fit_transient_both
from .misc import read_data

plt.rcParams['figure.figsize'] = (10, 5)

FITDRIVER = {'decay': fit_transient_exp, 'raise': fit_transient_raise,
             'osc': fit_transient_dmp_osc, 'both': fit_transient_both}


def save_TransientResult_txt(result: TransientResult, dirname: str):
    '''
    save fitting result to the text file

    Args:
     result: fitting result
     dirname: name of the directory in which text files for fitting result are saved.
     name_of_dset: name of each data sets. If `name_of_dset` is None then it is set to [1,2,3,....]
     t: sequence of scan range
     eps: sequence of estimated error of each datasets

    Returns:
     `fit_summary.txt`: Summary for the fitting result
     `weight_{name_of_dset[i]}.txt`: Weight of each model component of i th dataset
     `phase_{name_of_dset[i]}.txt`: Phase factor of each oscillation component of i th dataset
     `fit_{name_of_dset[i]}.txt`: fitting curve for i th dataset
     `fit_osc_{name_of_dset[i]}.txt`: oscillation part of fitting curve for i th dataset [model = 'both']
     `fit_decay_{name_of_dset[i]}.txt`: decay part of fitting curve for i th dataset [model = 'both']
     `res_{name_f_dset[i]}_j.txt`: residual (fit-data) curve for j th scan of i th data
                                   The format of text file is (t, res, eps)


    Note:
     If `dirname` directory is not exists, it creates `dirname` directory.
    '''
    if not (Path.cwd()/dirname).exists():
        os.mkdir(dirname)

    with open(f'{dirname}/fit_summary.txt', 'w', encoding='utf-8') as f:
        f.write(str(result))

    for i in range(len(result['t'])):
        coeff_fmt = result['eps'][i].shape[1]*['%.8e']
        fit_fmt = (1+result['eps'][i].shape[1])*['%.8e']
        coeff_header_lst = []
        fit_header_lst = ['time_delay']
        for j in range(result['eps'][i].shape[1]):
            res_save = np.vstack(
                (result['t'][i], result['res'][i][:, j], result['eps'][i][:, j])).T
            np.savetxt(f"{dirname}/res_{result['name_of_dset'][i]}_{j+1}.txt", res_save,
                       fmt=['%.8e', '%.8e', '%.8e'],
                       header=f"time_delay \t res_{result['name_of_dset'][i]}_{j+1} \t eps")
            fit_header_lst.append(f"fit_{result['name_of_dset'][i]}_{j+1}")
            coeff_header_lst.append(f"tscan_{result['name_of_dset'][i]}_{j+1}")

        fit_header = '\t'.join(fit_header_lst)
        coeff_header = '\t'.join(coeff_header_lst)

        np.savetxt(f"{dirname}/weight_{result['name_of_dset'][i]}.txt", result['c'][i], fmt=coeff_fmt,
                   header=coeff_header)
        if result['model'] in ['dmp_osc', 'both']:
            np.savetxt(f"{dirname}/phase_{result['name_of_dset'][i]}.txt", result['phase'][i], fmt=coeff_fmt,
                       header=coeff_header)
        fit_save = np.vstack((result['t'][i], result['fit'][i].T)).T
        np.savetxt(f"{dirname}/fit_{result['name_of_dset'][i]}.txt",
                   fit_save, fmt=fit_fmt, header=fit_header)
        if result['model'] == 'both':
            fit_decay_save = np.vstack(
                (result['t'][i], result['fit_decay'][i].T)).T
            np.savetxt(f"{dirname}/fit_decay_{result['name_of_dset'][i]}.txt",
                       fit_decay_save, fmt=fit_fmt, header=fit_header)
            fit_osc_save = np.vstack(
                (result['t'][i], result['fit_osc'][i].T)).T
            np.savetxt(
                f"{dirname}/fit_osc_{result['name_of_dset'][i]}.txt", fit_osc_save, fmt=fit_fmt, header=fit_header)



def plot_TransientResult(result: TransientResult, same_t0: bool, save_fig: Optional[str] = None):
    '''
    plot fitting Result

    Args:
     result: fitting result
     save_fig: prefix of saved png plots. If `save_fig` is `None`, plots are displayed instead of being saved.
    '''

    start = 0
    t0_idx = 1+1*(result['irf'] == 'pv')
    for i in range(len(result['t'])):
        if same_t0:
            t0 = result['x'][t0_idx+start]
        for j in range(result['intensity'][i].shape[1]):
            if not same_t0:
                t0 = result['x'][t0_idx+start+j]
            fig = plt.figure(start+j+1)
            title = f'{result["name_of_dset"][i]} scan #{j+1}'
            subtitle = f"Chi squared: {result['red_chi2_ind'][i][j]: .2f}"
            plt.suptitle(title)
            sub1 = fig.add_subplot(221)
            sub1.set_title(subtitle)
            sub1.errorbar(result['t'][i], result['intensity'][i][:, j], result['eps'][i][:, j], marker='o', mfc='none',
                          label=f'expt {title}', linestyle='none', color='black')
            sub1.plot(result['t'][i], result['fit'][i][:, j],
                      label=f'fit {title}', color='red')
            sub1.legend()
            sub2 = fig.add_subplot(222)
            sub2.set_title(subtitle)
            sub2.errorbar(result['t'][i], result['intensity'][i][:, j], result['eps'][i][:, j], marker='o', mfc='none',
                          label=f'expt {title}', linestyle='none', color='black')
            sub2.plot(result['t'][i], result['fit'][i][:, j],
                      label=f'fit {title}', color='red')
            sub2.legend()
            sub2.set_xlim(-10*result['fwhm']+t0, 20*result['fwhm']+t0)
            sub3 = fig.add_subplot(223)
            if result['model'] in ['decay', 'raise', 'osc']:
                sub3.errorbar(result['t'][i], result['res'][i][:, j],
                              result['eps'][i][:, j], marker='o', mfc='none',
                              label=f'res {title}', linestyle='none', color='black')
            else:
                sub3.errorbar(result['t'][i], result['intensity'][i][:, j]-result['fit_decay'][i][:, j],
                              result['eps'][i][:, j], marker='o', mfc='none', label=f'expt osc {title}',
                              linestyle='none', color='black')
                sub3.plot(result['t'][i], result['fit_osc'][i][:, j],
                          label=f'fit osc {title}', color='red')
            sub3.legend()

            sub4 = fig.add_subplot(224)
            if result['model'] in ['decay', 'raise', 'osc']:
                sub4.errorbar(result['t'][i], result['res'][i][:, j],
                              result['eps'][i][:, j], marker='o', mfc='none',
                              label=f'res {title}', linestyle='none', color='black')
            else:
                sub4.errorbar(result['t'][i], result['intensity'][i][:, j]-result['fit_decay'][i][:, j],
                              result['eps'][i][:, j], marker='o', mfc='none',
                              label=f'expt osc {title}', linestyle='none', color='black')
                sub4.plot(result['t'][i], result['fit_osc'][i][:, j],
                          label=f'fit osc {title}', color='red')
            sub4.set_xlim(-10*result['fwhm']+t0, 20*result['fwhm']+t0)
            sub4.legend()
            if save_fig is not None:
                plt.savefig(
                    f'./{save_fig}/{result["name_of_dset"][i]}_{j+1}.png')
        start = start + result['intensity'][i].shape[1]
    if save_fig is None:
        plt.show()


description = '''
fit tscan: fitting experimental time trace spectrum data with the convolution of the sum of
1. exponential decay (mode = decay)
2. raise model (mode = raise)
3. damped oscillation (mode = osc)
4. exponential decay, damped oscillation (mode=both)
and irf function
There are three types of irf function (gaussian, cauchy, pseudo voigt)
To calculate the contribution of each life time component, it solve least linear square problem via scipy linalg lstsq module.
'''

epilog = '''
*Note
1. The number of time zero parameter should be same as the total number of scan to fit.
2. However if you set `same_t0` then number of time zero parameter should be same as the total number of dataset
3. Every scan file whose prefix of filename is same should have same scan range
4. if you set shape of irf to pseudo voigt (pv), then you should provide two full width at half maximum
   value for gaussian and cauchy parts, respectively.
5. If you did not set tau and `mode=decay` then `--no_base` option is discouraged.
6. If you set `mode=decay or raise` then any parameter whose subscript is `osc` is discarded (i.e. tau_osc, period_osc).
7. If you set `mode=osc` then `tau` parameter is discarded. Also, baseline feature is not included in fitting function.
8. The number of tau_osc and period_osc parameter should be same
9. If you set `mode=both` then you should set `tau`, `tau_osc` and `period_osc`. However the number of `tau` and `tau_osc` need not to be same.
'''

mode_help = '''
Mode of fitting
 `decay`: fitting with the sum of the convolution of exponential decay and instrumental response function
 `raise` : fitting with the sum of the convolution of raise model and instrumental response function
 `osc`: fitting with the sum of the convolution of damped oscillation and instrumental response function
 `both`: fitting with the sum of both decay and osc
'''

irf_help = '''
shape of instrument response functon
g: gaussian distribution
c: cauchy distribution
pv: pseudo voigt profile, linear combination of gaussian distribution and cauchy distribution
    pv = eta*c+(1-eta)*g
    the uniform fwhm parameter and
    mixing parameter are determined according to Journal of Applied Crystallography. 33 (6): 1311–1316.
'''

fwhm_G_help = '''
full width at half maximum for gaussian shape
It would not be used when you set cauchy irf function
'''

fwhm_L_help = '''
full width at half maximum for cauchy shape
It would not be used when you did not set irf or use gaussian irf function
'''

method_glb_help = '''
Global optimization Method.
* 'basinhopping' : basinhopping
* 'ampgo' : adaptive memory programming for global optimization
If method_glb is not set, global optimization algorithm is not used.
'''


def fit_tscan():

    tmp = argparse.RawTextHelpFormatter
    parser = argparse.ArgumentParser(formatter_class=tmp,
                                     description=description,
                                     epilog=epilog)
    parser.add_argument('--mode', default='decay', choices=['decay', 'raise', 'osc', 'both'],
                        help=mode_help)
    parser.add_argument('--irf', default='g', choices=['g', 'c', 'pv'],
                        help=irf_help)
    parser.add_argument('--fwhm_G', type=float,
                        help=fwhm_G_help)
    parser.add_argument('--fwhm_L', type=float,
                        help=fwhm_L_help)
    parser.add_argument('prefix', nargs='+',
                        help='prefix for tscan files ' +
                        'It will read prefix_i.txt')
    parser.add_argument('--num_file', type=int, nargs='+',
                        help='number of scan file corresponding to each prefix')
    parser.add_argument('-t0', '--time_zeros', type=float, nargs='+',
                        help='time zeros for each tscan')
    parser.add_argument('-t0f', '--time_zeros_file',
                        help='filename for time zeros of each tscan')
    parser.add_argument('--tau', type=float, nargs='*',
                        help='lifetime of each decay component [mode: decay, both]')
    parser.add_argument('--tau_osc', type=float, nargs='+',
                        help='lifetime of each damped oscillation component [mode: osc, both]')
    parser.add_argument('--period_osc', type=float, nargs='+',
                        help='period of the vibration of each damped oscillation component [mode: osc, both]')
    parser.add_argument('--no_base', action='store_false',
                        help='exclude baseline for fitting [mode: decay, both]')
    parser.add_argument('--same_t0', action='store_true',
                        help='set time zero of every time delay scan belong to the same dataset same')
    parser.add_argument('--fix_irf', action='store_true',
                        help='fix irf parameter (fwhm_G, fwhm_L) during fitting process')
    parser.add_argument('--fix_t0', action='store_true',
                        help='fix time zero parameter during fitting process.')
    parser.add_argument('--fix_raise', action='store_true',
                        help='fix raise time constant [mode: raise]')
    parser.add_argument('--method_glb', choices=['basinhopping', 'ampgo'],
                        help=method_glb_help)
    parser.add_argument('-o', '--outdir', default='out',
                        help='name of directory to store output files')
    parser.add_argument('--save_fig', action='store_true',
                        help='save plot instead of display')
    args = parser.parse_args()

    prefix = np.array(args.prefix, dtype=str)
    num_file = np.array(args.num_file, dtype=int)

    irf = args.irf
    if irf == 'g':
        if args.fwhm_G is None:
            raise Exception(
                'You are using gaussian irf, so you should set fwhm_G!\n')
        else:
            fwhm_init = args.fwhm_G
    elif irf == 'c':
        if args.fwhm_L is None:
            raise Exception('You are using cauchy/lorenzian irf,' +
                            'so you should set fwhm_L!\n')
        else:
            fwhm_init = args.fwhm_L
    else:
        if (args.fwhm_G is None) or (args.fwhm_L is None):
            raise Exception('You are using pseudo voigt irf,' +
                            'so you should set both fwhm_G and fwhm_L!\n')
        else:
            fwhm_init = np.array([args.fwhm_G, args.fwhm_L])

    if (args.time_zeros is None) and (args.time_zeros_file is None):
        raise Exception(
            'You should set either time_zeros or time_zeros_file!\n')

    elif args.time_zeros is None:
        t0_init = np.genfromtxt(args.time_zeros_file)
    else:
        t0_init = np.array(args.time_zeros)

    t = np.empty(prefix.size, dtype=object)
    intensity = np.empty(prefix.size, dtype=object)
    eps = np.empty(prefix.size, dtype=object)
    num_scan = np.sum(num_file)

    if (not args.same_t0) and (num_scan != t0_init.size):
        raise Exception(
            'the Number of initial time zero parameter should be same as num_file parameter')
    elif args.same_t0 and (t0_init.size != len(num_file)):
        raise Exception('You set `same_t0`,'+
        ' so the number of initial time zero paramter should be same as the number of prefix argument')

    for i in range(prefix.size):
        t[i] = np.genfromtxt(f'{prefix[i]}_1.txt')[:, 0]
        num_data_pts = t[i].size
        intensity[i], eps[i] = read_data(
            prefix[i], num_file[i], num_data_pts, 10)

    print(f'fitting with total {num_scan} data set!\n')

    bound_fwhm = None
    if args.fix_irf:
        if irf in ['g', 'c']:
            bound_fwhm = [(fwhm_init, fwhm_init)]
        else:
            bound_fwhm = [(fwhm_init[0], fwhm_init[0]),
                          (fwhm_init[1], fwhm_init[1])]

    bound_t0 = None
    if args.fix_t0:
        bound_t0 = t0_init.size*[None]
        for i in range(t0_init.size):
            bound_t0[i] = (t0_init[i], t0_init[i])

    dargs = []
    base = False
    if args.mode in ['decay', 'raise', 'both']:
        if args.tau is None:
            base = True
            tau_init = None
        else:
            tau_init = np.array(args.tau)
            base = args.no_base
        dargs.append(tau_init)
        if args.mode in ['decay', 'raise']:
            dargs.append(base)
    if args.mode in ['osc', 'both']:
        tau_osc_init = np.array(args.tau_osc)
        period_osc_init = np.array(args.period_osc)
        dargs.append(tau_osc_init)
        dargs.append(period_osc_init)

    if args.mode == 'both':
        dargs.append(base)

    bound_tau = None
    if args.fix_raise and args.mode == 'raise':
        bound_tau = []
        for tau in tau_init:
            bound_tau.append(set_bound_tau(tau, fwhm_init))
        bound_tau[0] = (tau_init[0], tau_init[0])

    result = FITDRIVER[args.mode](irf, fwhm_init, t0_init, *dargs, method_glb=args.method_glb,
                                  bound_fwhm=bound_fwhm, bound_t0=bound_t0, bound_tau=bound_tau,
                                  same_t0=args.same_t0,
                                  name_of_dset=prefix, t=t, intensity=intensity, eps=eps)

    save_TransientResult(result, args.outdir)
    save_TransientResult_txt(result, args.outdir)
    print(result)
    if args.save_fig:
        plot_TransientResult(result, args.same_t0, args.outdir)
    else:
        plot_TransientResult(result, args.same_t0)




# pylint: disable = missing-module-docstring, wrong-import-position
# fit tscan gui py
# Wrapper script for fit_tscan_gui()
# Date: 2022. 12. 20.
# Author: pistack
# Email: pistack@yonsei.ac.kr

import os
import sys
path = os.path.dirname(os.path.realpath(__file__))
sys.path.append(path+"/../src/")
from TRXASprefitpack.tools._fit_tscan_gui import fit_tscan_gui

if __name__ == '__main__':
    fit_tscan_gui()



'''
res_decay:
submodule for residual function and gradient for fitting time delay scan with the
convolution of sum of exponential decay and instrumental response function

:copyright: 2021-2022 by pistack (Junho Lee).
:license: LGPL3.
'''

from typing import Optional, Sequence, Tuple
import numpy as np
from ..mathfun.irf import calc_eta, deriv_eta
from ..mathfun.irf import calc_fwhm, deriv_fwhm
from ..mathfun.A_matrix import make_A_matrix_gau, make_A_matrix_cauchy, fact_anal_A
from ..mathfun.exp_conv_irf import deriv_exp_conv_gau, deriv_exp_conv_cauchy
from ..mathfun.exp_conv_irf import deriv_exp_sum_conv_gau, deriv_exp_sum_conv_cauchy

# residual and gradient function for exponential decay model


def residual_decay(x0: np.ndarray, base: bool, irf: str,
                   t: Optional[Sequence[np.ndarray]] = None,
                   intensity: Optional[Sequence[np.ndarray]] = None, eps: Optional[Sequence[np.ndarray]] = None) -> np.ndarray:
    '''
    residual_decay
    scipy.optimize.least_squares compatible vector residual function for fitting multiple set of time delay scan with the
    sum of convolution of exponential decay and instrumental response function

    Args:
     x0: initial parameter,
      if irf == 'g','c':

        * 1st: fwhm_(G/L)
        * 2nd to :math:`2+N_{scan}`: time zero of each scan
        * :math:`2+N_{scan}` to :math:`2+N_{scan}+N_{\\tau}`: time constant of each decay component

      if irf == 'pv':

        * 1st and 2nd: fwhm_G, fwhm_L
        * 3rd to :math:`3+N_{scan}`: time zero of each scan
        * :math:`3+N_{scan}` to :math:`3+N_{scan}+N_{\\tau}`: time constant of each decay component

     num_comp: number of exponential decay component (except base)
     base: whether or not include baseline (i.e. very long lifetime component)
     irf: shape of instrumental response function

          * 'g': normalized gaussian distribution,
          * 'c': normalized cauchy distribution,
          * 'pv': pseudo voigt profile :math:`(1-\\eta)g(f) + \\eta c(f)`

        For pseudo voigt profile, the mixing parameter :math:`\\eta(f_G, f_L)` and
        uniform fwhm paramter :math:`f(f_G, f_L)` are calculated by `calc_eta` and `calc_fwhm` routine
     t: time points for each data set
     intensity: sequence of intensity of datasets
     eps: sequence of estimated error of datasets

    Returns:
     Residual vector
    '''

    x0 = np.atleast_1d(x0)

    if irf in ['g', 'c']:
        num_irf = 1
        fwhm = x0[0]
    else:
        num_irf = 2
        fwhm = calc_fwhm(x0[0], x0[1])
        eta = calc_eta(x0[0], x0[1])

    num_t0 = 0
    count = 0
    for d in intensity:
        num_t0 = d.shape[1] + num_t0
        count = count + d.size

    chi = np.empty(count)
    tau = x0[num_irf+num_t0:]
    if not base:
        k = 1/tau
    else:
        k = np.empty(tau.size+1)
        k[:-1] = 1/tau
        k[-1] = 0

    end = 0
    t0_idx = num_irf
    for ti, d, e in zip(t, intensity, eps):
        for j in range(d.shape[1]):
            t0 = x0[t0_idx]
            if irf == 'g':
                A = make_A_matrix_gau(ti-t0, fwhm, k)
            elif irf == 'c':
                A = make_A_matrix_cauchy(ti-t0, fwhm, k)
            else:
                A_gau = make_A_matrix_gau(ti-t0, fwhm, k)
                A_cauchy = make_A_matrix_cauchy(ti-t0, fwhm, k)
                A = A_gau + eta*(A_cauchy-A_gau)
            c = fact_anal_A(A, d[:, j], e[:, j])
            chi[end:end+d.shape[0]] = ((c@A) - d[:, j])/e[:, j]

            end = end + d.shape[0]
            t0_idx = t0_idx + 1

    return chi


def res_grad_decay(x0: np.ndarray, num_comp: int, base: bool, irf: str,
                   fix_param_idx: Optional[np.ndarray] = None,
                   t: Optional[Sequence[np.ndarray]] = None,
                   intensity: Optional[Sequence[np.ndarray]] = None,
                   eps: Optional[Sequence[np.ndarray]] = None) -> Tuple[np.ndarray, np.ndarray]:
    '''
    res_grad_decay
    scipy.optimize.minimize compatible scalar residual and its gradient function for fitting multiple set of time delay scan with the
    sum of convolution of exponential decay and instrumental response function

    Args:
     x0: initial parameter,
      if irf == 'g','c':

        * 1st: fwhm_(G/L)
        * 2nd to :math:`2+N_{scan}`: time zero of each scan
        * :math:`2+N_{scan}` to :math:`2+N_{scan}+N_{\\tau}`: time constant of each decay component

      if irf == 'pv':

        * 1st and 2nd: fwhm_G, fwhm_L
        * 3rd to :math:`3+N_{scan}`: time zero of each scan
        * :math:`3+N_{scan}` to :math:`3+N_{scan}+N_{\\tau}`: time constant of each decay component

     num_comp: number of exponential decay component (except base)
     base: whether or not include baseline (i.e. very long lifetime component)
     irf: shape of instrumental response function

          * 'g': normalized gaussian distribution,
          * 'c': normalized cauchy distribution,
          * 'pv': pseudo voigt profile :math:`(1-\\eta)g(f) + \\eta c(f)`

        For pseudo voigt profile, the mixing parameter :math:`\\eta(f_G, f_L)` and
        uniform fwhm paramter :math:`f(f_G, f_L)` are calculated by `calc_eta` and `calc_fwhm` routine
     t: time points for each data set
     fix_param_idx: index for fixed parameter (masked array for `x0`)
     intensity: sequence of intensity of datasets
     eps: sequence of estimated error of datasets

    Returns:
     Tuple of scalar residual function :math:`(\\frac{1}{2}\\sum_i {res}^2_i)` and its gradient
    '''
    x0 = np.atleast_1d(x0)

    if irf in ['g', 'c']:
        num_irf = 1
        fwhm = x0[0]
    else:
        num_irf = 2
        eta = calc_eta(x0[0], x0[1])
        fwhm = calc_fwhm(x0[0], x0[1])
        dfwhm_G, dfwhm_L = deriv_fwhm(x0[0], x0[1])
        deta_G, deta_L = deriv_eta(x0[0], x0[1])

    num_t0 = 0
    count = 0
    for d in intensity:
        num_t0 = num_t0 + d.shape[1]
        count = count + d.size

    tau = x0[num_irf+num_t0:]

    if not base:
        k = 1/tau
    else:
        k = np.empty(tau.size+1)
        k[:-1] = 1/tau
        k[-1] = 0

    num_param = num_irf+num_t0+num_comp
    chi = np.empty(count)
    df = np.empty((count, tau.size+num_irf))
    grad = np.empty(num_param)

    end = 0
    t0_idx = num_irf

    for ti, d, e in zip(t, intensity, eps):
        step = d.shape[0]
        for j in range(d.shape[1]):
            t0 = x0[t0_idx]
            if irf == 'g':
                A = make_A_matrix_gau(ti-t0, fwhm, k)
            elif irf == 'c':
                A = make_A_matrix_cauchy(ti-t0, fwhm, k)
            else:
                A_gau = make_A_matrix_gau(ti-t0, fwhm, k)
                A_cauchy = make_A_matrix_cauchy(ti-t0, fwhm, k)
                diff = A_cauchy-A_gau
                A = A_gau + eta*diff
            c = fact_anal_A(A, d[:, j], e[:, j])
            chi[end:end+step] = (c@A-d[:, j])/e[:, j]

            if irf == 'g':
                grad_tmp = deriv_exp_sum_conv_gau(ti-t0, fwhm, 1/tau, c, base)
            elif irf == 'c':
                grad_tmp = deriv_exp_sum_conv_cauchy(
                    ti-t0, fwhm, 1/tau, c, base)
            else:
                grad_gau = deriv_exp_sum_conv_gau(ti-t0, fwhm, 1/tau, c, base)
                grad_cauchy = deriv_exp_sum_conv_cauchy(
                    ti-t0, fwhm, 1/tau, c, base)
                grad_tmp = grad_gau + eta*(grad_cauchy-grad_gau)

            grad_tmp = np.einsum('i,ij->ij', 1/e[:, j], grad_tmp)
            if irf in ['g', 'c']:
                df[end:end+step, 0] = grad_tmp[:, 1]
            else:
                cdiff = (c@diff)/e[:, j]
                df[end:end+step, 0] = dfwhm_G*grad_tmp[:, 1]+deta_G*cdiff
                df[end:end+step, 1] = dfwhm_L*grad_tmp[:, 1]+deta_L*cdiff
            grad[t0_idx] = -chi[end:end+step]@grad_tmp[:, 0]
            df[end:end+step,
                num_irf:] = np.einsum('j,ij->ij', -1/tau**2, grad_tmp[:, 2:])

            end = end + step
            t0_idx = t0_idx + 1

    mask = np.ones(num_param, dtype=bool)
    mask[num_irf:num_irf+num_t0] = False
    grad[mask] = chi@df

    if fix_param_idx is not None:
        grad[fix_param_idx] = 0

    return np.sum(chi**2)/2, grad

def residual_decay_same_t0(x0: np.ndarray, base: bool, irf: str,
                           t: Optional[Sequence[np.ndarray]] = None,
                           intensity: Optional[Sequence[np.ndarray]] = None,
                           eps: Optional[Sequence[np.ndarray]] = None) -> np.ndarray:
    '''
    residual_decay_same_t0
    scipy.optimize.least_squares compatible vector residual function
    for fitting multiple set of time delay scan with the
    sum of convolution of exponential decay and instrumental response function
    Set Time Zero of every time dset in same dataset same

    Args:
     x0: initial parameter,
      if irf == 'g','c':

        * 1st: fwhm_(G/L)
        * 2nd to :math:`2+N_{dset}`: time zero of each data set
        * :math:`2+N_{dset}` to :math:`2+N_{dset}+N_{\\tau}`: time constant of each decay component

      if irf == 'pv':

        * 1st and 2nd: fwhm_G, fwhm_L
        * 3rd to :math:`3+N_{dset}`: time zero of each data set
        * :math:`3+N_{dset}` to :math:`3+N_{dset}+N_{\\tau}`: time constant of each decay component

     num_comp: number of exponential decay component (except base)
     base: whether or not include baseline (i.e. very long lifetime component)
     irf: shape of instrumental response function

          * 'g': normalized gaussian distribution,
          * 'c': normalized cauchy distribution,
          * 'pv': pseudo voigt profile :math:`(1-\\eta)g(f) + \\eta c(f)`

        For pseudo voigt profile, the mixing parameter :math:`\\eta(f_G, f_L)` and
        uniform fwhm paramter :math:`f(f_G, f_L)` are calculated by `calc_eta` and `calc_fwhm` routine
     t: time points for each data set
     intensity: sequence of intensity of datasets
     eps: sequence of estimated error of datasets

    Returns:
     Residual vector
    '''

    x0 = np.atleast_1d(x0)

    if irf in ['g', 'c']:
        num_irf = 1
        fwhm = x0[0]
    else:
        num_irf = 2
        fwhm = calc_fwhm(x0[0], x0[1])
        eta = calc_eta(x0[0], x0[1])

    num_dataset = len(t)
    count = 0
    for i in range(num_dataset):
        count = count + intensity[i].size

    chi = np.empty(count)
    tau = x0[num_irf+num_dataset:]
    if not base:
        k = 1/tau
    else:
        k = np.empty(tau.size+1)
        k[:-1] = 1/tau
        k[-1] = 0

    end = 0
    t0_idx = num_irf
    for ti, d, e in zip(t, intensity, eps):
        t0 = x0[t0_idx]
        if irf == 'g':
            A = make_A_matrix_gau(ti-t0, fwhm, k)
        elif irf == 'c':
            A = make_A_matrix_cauchy(ti-t0, fwhm, k)
        else:
            A_gau = make_A_matrix_gau(ti-t0, fwhm, k)
            A_cauchy = make_A_matrix_cauchy(ti-t0, fwhm, k)
            A = A_gau + eta*(A_cauchy-A_gau)
        for j in range(d.shape[1]):
            c = fact_anal_A(A, d[:, j], e[:, j])
            chi[end:end+d.shape[0]] = ((c@A) - d[:, j])/e[:, j]

            end = end + d.shape[0]
        t0_idx = t0_idx + 1

    return chi

def res_grad_decay_same_t0(x0: np.ndarray, num_comp: int, base: bool, irf: str,
                           fix_param_idx: Optional[np.ndarray] = None,
                           t: Optional[Sequence[np.ndarray]] = None,
                           intensity: Optional[Sequence[np.ndarray]] = None,
                           eps: Optional[Sequence[np.ndarray]] = None) -> Tuple[np.ndarray, np.ndarray]:
    '''
    res_grad_decay_same_t0
    scipy.optimize.minimize compatible scalar residual
    and its gradient function for fitting multiple set of time delay scan with the
    sum of convolution of exponential decay and instrumental response function

    Args:
     x0: initial parameter,
      if irf == 'g','c':

        * 1st: fwhm_(G/L)
        * 2nd to :math:`2+N_{dset}`: time zero of each dataset
        * :math:`2+N_{dset}` to :math:`2+N_{dset}+N_{\\tau}`: time constant of each decay component

      if irf == 'pv':

        * 1st and 2nd: fwhm_G, fwhm_L
        * 3rd to :math:`3+N_{dset}`: time zero of each dataset
        * :math:`3+N_{dset}` to :math:`3+N_{dset}+N_{\\tau}`: time constant of each decay component

     num_comp: number of exponential decay component (except base)
     base: whether or not include baseline (i.e. very long lifetime component)
     irf: shape of instrumental response function

          * 'g': normalized gaussian distribution,
          * 'c': normalized cauchy distribution,
          * 'pv': pseudo voigt profile :math:`(1-\\eta)g(f) + \\eta c(f)`

        For pseudo voigt profile, the mixing parameter :math:`\\eta(f_G, f_L)` and
        uniform fwhm paramter :math:`f(f_G, f_L)` are calculated by `calc_eta` and `calc_fwhm` routine
     t: time points for each data set
     fix_param_idx: index for fixed parameter (masked array for `x0`)
     intensity: sequence of intensity of datasets
     eps: sequence of estimated error of datasets

    Returns:
     Tuple of scalar residual function :math:`(\\frac{1}{2}\\sum_i {res}^2_i)` and its gradient
    '''

    x0 = np.atleast_1d(x0)

    if irf in ['g', 'c']:
        num_irf = 1
        fwhm = x0[0]
        eta = None
    else:
        num_irf = 2
        eta = calc_eta(x0[0], x0[1])
        fwhm = calc_fwhm(x0[0], x0[1])
        deta_G, deta_L = deriv_eta(x0[0], x0[1])
        dfwhm_G, dfwhm_L = deriv_fwhm(x0[0], x0[1])

    num_dataset = len(t)
    count = 0
    for i in range(num_dataset):
        count = count + intensity[i].size

    tau = x0[num_irf+num_dataset:num_irf+num_dataset+num_comp]

    if base:
        k = np.empty(tau.size+1)
        k[:-1] = 1/tau
        k[-1] = 0
    else:
        k = 1/tau

    num_param = num_irf+num_dataset+num_comp

    chi = np.empty(count)
    df = np.zeros((count, num_irf+num_comp))
    grad = np.zeros(num_param)

    end = 0
    t0_idx = num_irf
    for ti, d, e in zip(t, intensity, eps):
        step = d.shape[0]
        A = np.empty((num_comp+1*base, step))
        A_grad_decay = np.empty((num_comp+1*base, step, 3))
        grad_decay = np.empty((step, num_comp+2))

        t0 = x0[t0_idx]
        if irf == 'g':
            A[:num_comp+1*base, :] = make_A_matrix_gau(ti-t0, fwhm, k)
            for i in range(num_comp):
                A_grad_decay[i, :, :] = deriv_exp_conv_gau(ti-t0, fwhm, k[i])
            if base:
                A_grad_decay[-1, :, :] = deriv_exp_conv_gau(ti-t0, fwhm, 0)
        elif irf == 'c':
            A[:num_comp+1*base, :] = make_A_matrix_cauchy(ti-t0, fwhm, k)
            for i in range(num_comp):
                A_grad_decay[i, :, :] = deriv_exp_conv_cauchy(ti-t0, fwhm, k[i])
            if base:
                A_grad_decay[-1, :, :] = deriv_exp_conv_cauchy(ti-t0, fwhm, 0)
        else:
            tmp_gau = make_A_matrix_gau(ti-t0, fwhm, k)
            tmp_cauchy = make_A_matrix_cauchy(ti-t0, fwhm, k)
            diff = tmp_cauchy-tmp_gau
            A[:num_comp+1*base, :] = tmp_gau + eta*diff
            for i in range(num_comp):
                tmp_grad_gau = deriv_exp_conv_gau(ti-t0, fwhm, k[i])
                tmp_grad_cauchy = deriv_exp_conv_cauchy(ti-t0, fwhm, k[i])
                A_grad_decay[i, :, :] = tmp_grad_gau + eta*(tmp_grad_cauchy-tmp_grad_gau)
            if base:
                tmp_grad_gau = deriv_exp_conv_gau(ti-t0, fwhm, 0)
                tmp_grad_cauchy = deriv_exp_conv_cauchy(ti-t0, fwhm, 0)
                A_grad_decay[-1, :, :] = tmp_grad_gau + eta*(tmp_grad_cauchy-tmp_grad_gau)

        for j in range(d.shape[1]):
            c = fact_anal_A(A, d[:, j], e[:, j])
            chi[end:end+step] = (c@A-d[:, j])/e[:, j]

            grad_decay[:, :2] = \
                np.tensordot(c[:num_comp+1*base], A_grad_decay[:, :, :2], axes=1)
            for i in range(num_comp):
                grad_decay[:, 2+i] = c[i]*A_grad_decay[i, :, 2]

            grad_decay = np.einsum('i,ij->ij', 1/e[:, j], grad_decay)

            if irf in ['g', 'c']:
                df[end:end+step, 0] = grad_decay[:, 1]

            else:
                cdiff = (c[:num_comp+1*base]@diff)/e[:, j]
                df[end:end+step, 0] = dfwhm_G*grad_decay[:, 1]+deta_G*cdiff
                df[end:end+step, 1] = dfwhm_L*grad_decay[:, 1]+deta_L*cdiff

            grad[t0_idx] = grad[t0_idx] -chi[end:end+step]@grad_decay[:, 0]
            df[end:end+step, num_irf:num_irf+num_comp] = \
                np.einsum('j,ij->ij', -1/tau**2, grad_decay[:, 2:])

            end = end + step

        t0_idx = t0_idx + 1

    mask = np.ones(num_param, dtype=bool)
    mask[num_irf:num_irf+num_dataset] = False
    grad[mask] = chi@df

    if fix_param_idx is not None:
        grad[fix_param_idx] = 0

    return np.sum(chi**2)/2, grad



'''
exp_conv_irf:
submodule for the mathematical functions for
irf (instrumental response function)

:copyright: 2022 by pistack (Junho Lee).
:license: LGPL3.
'''

from typing import Union, Tuple
import numpy as np


def gau_irf(t: Union[float, np.ndarray], fwhm: float) -> Union[float, np.ndarray]:
    '''
    Compute gaussian shape irf function

    Args:
      t: time
      fwhm: full width at half maximum of gaussian distribution

    Returns:
     normalized gaussian function.
    '''

    sigma = fwhm/(2*np.sqrt(2*np.log(2)))
    return np.exp(-t**2/(2*sigma**2))/(sigma*np.sqrt(2*np.pi))


def cauchy_irf(t: Union[float, np.ndarray], fwhm: float) -> Union[float, np.ndarray]:
    '''
    Compute lorenzian shape irf function

    Args:
      t: time
      fwhm: full width at half maximum of cauchy distribution

    Returns:
     normalized lorenzian function.
    '''

    gamma = fwhm/2
    return gamma/np.pi/(t**2+gamma**2)


def calc_fwhm(fwhm_G: float, fwhm_L: float) -> float:
    '''
    Calculate uniform fwhm of both gaussian and cauchy component of
    pseudo voigt profile with fwhm_G, fwhm_L based on
    Journal of Applied Crystallography. 33 (6): 1311–1316.

    Args:
      fwhm_G: full width at half maximum of gaussian part
      fwhm_L: full width at half maximum of lorenzian part
    Returns:
     uniform fwhm parameter of both gaussian and cauchy component of
     pseudo voigt profile
    '''
    f = fwhm_G**5+2.69269*fwhm_G**4*fwhm_L + \
        2.42843*fwhm_G**3*fwhm_L**2 + \
        4.47163*fwhm_G**2*fwhm_L**3 + \
        0.07842*fwhm_G*fwhm_L**4 + \
        fwhm_L**5
    return f**(1/5)


def calc_eta(fwhm_G: float, fwhm_L: float) -> float:
    '''
    Calculate eta of pseudo voigt profile with fwhm_G, fwhm_L based on
    Journal of Applied Crystallography. 33 (6): 1311–1316.

    Args:
      fwhm_G: full width at half maximum of gaussian part
      fwhm_L: full width at half maximum of lorenzian part
    Returns:
     maxing parameter eta
    '''
    f = fwhm_G**5+2.69269*fwhm_G**4*fwhm_L + \
        2.42843*fwhm_G**3*fwhm_L**2 + \
        4.47163*fwhm_G**2*fwhm_L**3 + \
        0.07842*fwhm_G*fwhm_L**4 + \
        fwhm_L**5
    f = f**(1/5)
    x = fwhm_L/f
    eta = 1.36603*x-0.47719*x**2+0.11116*x**3
    return eta


def pvoigt_irf(t: Union[float, np.ndarray], fwhm: float, eta: float) -> Union[float, np.ndarray]:
    '''
    Compute pseudo voight shape irf function
    (i.e. linear combination of gaussian and lorenzian function)

    Args:
      t: time
      fwhm: uniform full width at half paramter
      eta: mixing paramter
    Returns:
     pseudo voigt profile
    '''
    u = gau_irf(t, fwhm)
    v = cauchy_irf(t, fwhm)

    return u + eta*(v-u)


def deriv_fwhm(fwhm_G: float, fwhm_L: float) -> Tuple[float, float]:
    '''
    Computes gradient of uniform fwhm parameter of pseudo voigt approximation
    based on Journal of Applied Crystallography. 33 (6): 1311-1316.

    Args:
     fwhm_G: full width at half maximum of gaussian part
     fwhm_L: full width at half maximum of lorenzian part

    Returns:
     gradient of fwhm(fwhm_G, fwhm_L)
    '''
    f = fwhm_G**5+2.69269*fwhm_G**4*fwhm_L + \
        2.42843*fwhm_G**3*fwhm_L**2 + \
        4.47163*fwhm_G**2*fwhm_L**3 + \
        0.07842*fwhm_G*fwhm_L**4 + \
        fwhm_L**5
    df_fwhm_G = 5*fwhm_G**4+10.77076*fwhm_G**3*fwhm_L + \
        7.28529*fwhm_G**2*fwhm_L**2+8.94326*fwhm_G*fwhm_L**3 + \
        0.07842*fwhm_L**4
    df_fwhm_L = 5*fwhm_L**4 + 0.31368*fwhm_L**3*fwhm_G + \
        13.41489*fwhm_G**2*fwhm_L**2 + 4.85686*fwhm_L*fwhm_G**3 + \
        2.69269*fwhm_G**4
    return df_fwhm_G/f**(4/5)/5, df_fwhm_L/f**(4/5)/5


def deriv_eta(fwhm_G: float, fwhm_L: float) -> Tuple[float, float]:
    '''
    Calculate gradient of eta of pseudo voigt profile with fwhm_G, fwhm_L based on
    Journal of Applied Crystallography. 33 (6): 1311–1316.

    Args:
      fwhm_G: full width at half maximum of gaussian part
      fwhm_L: full width at half maximum of lorenzian part
    Returns:
     gradient of eta(fwhm_G, fwhm_L)
    '''
    f = fwhm_G**5+2.69269*fwhm_G**4*fwhm_L + \
        2.42843*fwhm_G**3*fwhm_L**2 + \
        4.47163*fwhm_G**2*fwhm_L**3 + \
        0.07842*fwhm_G*fwhm_L**4 + \
        fwhm_L**5
    g = f**(-1/5)
    x = fwhm_L*g
    df_fwhm_G = 5*fwhm_G**4+10.77076*fwhm_G**3*fwhm_L + \
        7.28529*fwhm_G**2*fwhm_L**2+8.94326*fwhm_G*fwhm_L**3 + \
        0.07842*fwhm_L**4
    df_fwhm_L = 5*fwhm_L**4 + 0.31368*fwhm_L**3*fwhm_G + \
        13.41489*fwhm_G**2*fwhm_L**2 + 4.85686*fwhm_L*fwhm_G**3 + \
        2.69269*fwhm_G**4
    dx_fwhm_G = -fwhm_L*df_fwhm_G*g/f/5
    dx_fwhm_L = g - fwhm_L*df_fwhm_L*g/f/5
    deta_x = 0.33348*x**2-0.95438*x+1.36603
    return deta_x*dx_fwhm_G, deta_x*dx_fwhm_L



# pylint: disable = missing-module-docstring, wrong-import-position
# calc dads gui py
# Wrapper script for fit_tscan_gui()
# Date: 2022. 12. 26.
# Author: pistack
# Email: pistack@yonsei.ac.kr

import os
import sys
path = os.path.dirname(os.path.realpath(__file__))
sys.path.append(path+"/../src/")
from TRXASprefitpack.tools._calc_dads_gui import calc_dads_gui

if __name__ == '__main__':
    calc_dads_gui()



'''
calc_dads_gui:
Graphical User Interface for calc_dads utility

:copyright: 2021-2022 by pistack (Junho Lee).
:license: LGPL3.
'''

from sys import platform
import re
import tkinter as tk
import tkinter.messagebox as msg
import tkinter.filedialog as fd
import numpy as np
from scipy.linalg import svd
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.backends.backend_tkagg import NavigationToolbar2Tk
from matplotlib.backend_bases import key_press_handler
from matplotlib.figure import Figure
from ..mathfun import calc_eta, calc_fwhm
from ..driver import dads

float_sep_comma = re.compile('([\+\-]?[0-9]+[.]?[0-9]*[,]\s*)*[\+\-]?[0-9]+[.]?[0-9]*\s*')
isfloat = re.compile('[\+\-]?[0-9]+[.]?[0-9]*\s*')

# check font
if platform == "linux" or platform == "linux2":
    widgetfont = 'Liberation Sans'
else:
    widgetfont = 'Arial'


class PlotDataWidget:
    '''
    Class for widget which plot data
    '''
    # These codes are based on Matplotlib example "Embedding in Tk"

    def __init__(self, master):

        self.top = tk.Toplevel(master.root)
        self.top.title('Plot Data')

        self.fig = Figure(figsize=(8, 4), dpi=100)

        # immutable
        self.ax = self.fig.add_subplot()
        self.ax.set_xlabel('Energy')
        self.ax.set_ylabel('Intensity')
        self.ax.grid(True)

        # mutable
        self.ax.set_title(f'{master.t[0]}')
        self.ln, = self.ax.plot(master.e, master.escan_mat[:, 0], mfc='none',
        color='black', marker='o')
        self.poly = self.ax.fill_between(master.e,
        master.escan_mat[:, 0]-master.eps[:, 0],
        master.escan_mat[:, 0]+master.eps[:, 0], alpha=0.5, color='black')

        if master.fit:
            self.ln_fit, = self.ax.plot(master.e, master.escan_fit[:, 0],
            color='red', marker=None)

        # set canvas
        self.canvas = FigureCanvasTkAgg(self.fig, master=self.top)
        self.canvas.draw()

        self.toolbar = \
            NavigationToolbar2Tk(self.canvas, self.top, pack_toolbar=False)
        self.toolbar.update()

        self.canvas.mpl_connect('key_press_event', key_press_handler)

        self.slider_update = tk.Scale(self.top, from_=1, to_=len(master.t),
        orient=tk.HORIZONTAL, command=lambda val: self.update_plot(master, int(val)),
        label='Time Delay index')

        self.slider_update.pack(side=tk.BOTTOM)
        self.toolbar.pack(side=tk.BOTTOM, fill=tk.X)
        self.canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=True)
        self.top.mainloop()

    # --- Update plot when one moves slider

    def update_plot(self, master, val):
        self.ax.set_title(f'{master.t[val-1]}')
        self.ln.set_data(master.e, master.escan_mat[:, val-1])
        self.poly.remove()
        self.poly = self.ax.fill_between(master.e,
        master.escan_mat[:, val-1]-master.eps[:, val-1],
        master.escan_mat[:, val-1]+master.eps[:, val-1], 
        alpha=0.5, color='black')
        if master.fit:
            self.ln_fit.set_data(master.e, master.escan_fit[:, val-1])

        ymin = np.min(master.escan_mat[:, val-1])
        ymax = np.max(master.escan_mat[:, val-1])
        errmax = np.max(master.eps[:, val-1])
        self.ax.set_ylim((ymin-5*errmax, ymax+5*errmax))
        self.canvas.draw()

class PlotSVDWidget:
    '''
    Class for widget which plot svd components
    '''
    # These codes are based on Matplotlib example "Embedding in Tk"

    def __init__(self, master):

        self.top = tk.Toplevel(master.root)
        self.top.title('Plot SVD result')

        self.fig = Figure(figsize=(12, 4), dpi=100)

        # immutable
        self.ax_e = self.fig.add_subplot(121)

        self.ax_e.set_xlabel('Energy')
        self.ax_e.set_ylabel('Intensity')
        self.ax_e.grid(True)
        self.ax_e.set_xlim((np.min(master.e), np.max(master.e)))
        self.ax_t = self.fig.add_subplot(122)
        self.ax_t.set_xlabel('Time Delay')
        self.ax_t.set_ylabel('Intensity')
        self.ax_t.grid(True)
        self.ax_t.set_xlim((np.min(master.t), np.max(master.t)))

        # mutable
        self.ax_e.set_title(f'Left Component: {1 :.3e}')
        self.ax_t.set_title(f'Right Component: {1 :.3e}')
        self.ln_e, = self.ax_e.plot(master.e, master.U[:, 0], mfc='none',
        color='black', marker='o')
        self.ln_t, = self.ax_t.plot(master.t, master.Vh[0, :], mfc='none',
        color='black', marker='o')

        # set canvas
        self.canvas = FigureCanvasTkAgg(self.fig, master=self.top)
        self.canvas.draw()

        self.toolbar = \
            NavigationToolbar2Tk(self.canvas, self.top, pack_toolbar=False)
        self.toolbar.update()

        self.canvas.mpl_connect('key_press_event', key_press_handler)

        self.slider_update = tk.Scale(self.top, from_=1, to_=len(master.sigma),
        orient=tk.HORIZONTAL, command=lambda val: self.update_plot(master, int(val)),
        label='Component index')

        self.slider_update.pack(side=tk.BOTTOM)
        self.toolbar.pack(side=tk.BOTTOM, fill=tk.X)
        self.canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=True)
        self.top.mainloop()

    # --- Update plot when one moves slider

    def update_plot(self, master, val):
        self.ax_e.set_title(f'Left Component: {master.sigma[val-1]/master.sigma[0] :.3e}')
        self.ax_t.set_title(f'Right Component: {master.sigma[val-1]/master.sigma[0] :.3e}')
        self.ln_e.set_data(master.e, master.U[:, val-1])
        self.ln_t.set_data(master.t, master.Vh[val-1, :])
        y_emin = np.min(master.U[:, val-1])
        y_emax = np.max(master.U[:, val-1])
        y_tmin = np.min(master.Vh[val-1, :])
        y_tmax = np.max(master.Vh[val-1, :])
        self.ax_e.set_ylim(((y_emin+y_emax)/2-1.05*(y_emax-y_emin)/2,
        (y_emin+y_emax)/2+1.05*(y_emax-y_emin)/2))
        self.ax_t.set_ylim(((y_tmin+y_tmax)/2-1.05*(y_tmax-y_tmin)/2,
        (y_tmin+y_tmax)/2+1.05*(y_tmax-y_tmin)/2))
        self.canvas.draw()


class PlotDADSWidget:
    '''
    Class for wideget which plots DADS
    '''
    # These codes are based on Matplotlib example "Embedding in Tk"

    def __init__(self, master):

        self.top = tk.Toplevel(master.root)
        self.top.title('Plot DADS')

        self.fig = Figure(figsize=(8, 4), dpi=100)

        self.ax_dads = self.fig.add_subplot()
        self.ax_dads.set_xlabel('Energy')
        self.ax_dads.set_ylabel('Intensity')
        self.ax_dads.grid(True)

        for i in range(len(master.tau)):
            self.ax_dads.plot(master.e, master.dads[:, i], label=f'tau: {master.tau[i]}')
        
        if master.base_var.get():
            self.ax_dads.plot(master.e, master.dads[:, -1], label='long lived')

        self.ax_dads.legend()

        # set canvas
        self.canvas = FigureCanvasTkAgg(self.fig, master=self.top)
        self.canvas.draw()

        self.toolbar = \
            NavigationToolbar2Tk(self.canvas, self.top, pack_toolbar=False)
        self.toolbar.update()

        self.canvas.mpl_connect('key_press_event', key_press_handler)
        self.toolbar.pack(side=tk.BOTTOM, fill=tk.X)
        self.canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=True)
        self.top.mainloop()

class CalcDADSGuiWidget:
    '''
    Class for calc_dads gui wrapper
    '''
    def __init__(self):
        self.root = tk.Tk()
        self.root.title('Calc DADS Gui')
        self.fit = False

        # -- define necessary variables
        self.irf_var = tk.StringVar()
        self.base_var = tk.IntVar()
        self.escan_file = None
        self.time_file = None
        self.eps_file = None

        # --- irf model selection window
        self.irf_label = tk.Label(self.root, text='Select Type of irf',
        padx=90, pady=10, font=(widgetfont, 12))
        self.irf_label.grid(column=0, row=2, columnspan=3)
        self.irf_g = tk.Checkbutton(self.root, text='gaussian', variable=self.irf_var,
        onvalue='g', offvalue='')
        self.irf_g.grid(column=0, row=3)
        self.irf_c = tk.Checkbutton(self.root, text='cauchy', variable=self.irf_var,
        onvalue='c', offvalue='')
        self.irf_c.grid(column=1, row=3)
        self.irf_pv = tk.Checkbutton(self.root, text='pseudo voigt', variable=self.irf_var,
        onvalue='pv', offvalue='')
        self.irf_pv.grid(column=2, row=3)

        # --- SVD options
        self.option_label = tk.Label(self.root, text='SVD Options',
        padx=90, pady=10, font=(widgetfont, 12))
        self.option_label.grid(column=0, row=4, columnspan=3)
        self.label_cond_num = tk.Label(self.root, text='Conditional Number',
        padx=60, pady=10, font=(widgetfont, 12))
        self.label_cond_num.grid(column=0, row=5, columnspan=2)
        self.entry_cond_num = tk.Entry(self.root, width=10)
        self.entry_cond_num.grid(column=2, row=5)
        self.entry_cond_num.insert(0, '0')

        # --- miscellaneous options
        self.option_label = tk.Label(self.root, text='Miscellaneous Options',
        padx=120, pady=10, font=(widgetfont, 12))
        self.option_label.grid(column=0, row=6, columnspan=4)
        self.include_base_check = tk.Checkbutton(self.root, text='base',
        variable=self.base_var, onvalue=1, offvalue=0)
        self.include_base_check.grid(column=0, row=7)

        # --- Read file to fit
        self.label_file = tk.Label(self.root, text='Browse Files',
        padx=120, pady=10, font=(widgetfont, 12))
        self.label_file.grid(column=0, row=8, columnspan=4)
        self.print_file_num = tk.Canvas(self.root, width=320, height=20, bd=5,
        bg='white')
        self.print_file_num.grid(column=0, row=9, columnspan=2)
        self.button_file = tk.Button(self.root, width=30, bd=5, text='browse',
        command=self.browse_file)
        self.button_file.grid(column=2, row=9)
        self.button_plot = tk.Button(self.root, width=30, bd=5, text='plot',
        command=self.plot_file)
        self.button_plot.grid(column=3, row=9)

        # --- Parameters
        # 1. fwhm of irf function
        # 2. time zero of energy scan
        # 3. time constant

        self.label_paramter = tk.Label(self.root, text='Parameter',
        font=(widgetfont, 15), padx=90, pady=10)
        self.label_paramter.grid(column=0, row=10, columnspan=3)
        self.label_fwhm_G = tk.Label(self.root, text='fwhm_G (irf)',
        padx=30, pady=10, font=(widgetfont, 12))
        self.label_fwhm_G.grid(column=0, row=11)
        self.entry_fwhm_G = tk.Entry(self.root, width=10, bd=1)
        self.entry_fwhm_G.grid(column=0, row=12)

        self.label_fwhm_L = tk.Label(self.root, text='fwhm_L (irf)',
        padx=30, pady=10, font=(widgetfont, 12))
        self.label_fwhm_L.grid(column=1, row=11)
        self.entry_fwhm_L = tk.Entry(self.root, width=10, bd=1)
        self.entry_fwhm_L.grid(column=1, row=12)

        self.label_t0 = tk.Label(self.root,
        text='t0 (escan)',
        padx=30, pady=10, font=(widgetfont, 12))
        self.label_t0.grid(column=2, row=11)
        self.entry_t0 = tk.Entry(self.root, width=10, bd=1)
        self.entry_t0.grid(column=2, row=12)

        self.label_tau = tk.Label(self.root,
        text='Insert life time parameter (tau1,tau2,...)',
        padx=90, pady=10, font=(widgetfont, 12))
        self.label_tau.grid(column=0, row=13, columnspan=3)
        self.entry_tau = tk.Entry(self.root, width=90, bd=5)
        self.entry_tau.grid(column=0, row=14, columnspan=3)

        # -- Buttons

        self.svd_button = tk.Button(self.root,
        text='SVD', command=self.svd_script,
        font=(widgetfont, 12), padx=30, pady=10, bd=5)
        self.svd_button.grid(column=0, row=16)

        self.run_button = tk.Button(self.root,
        text='Run', command=self.run_script,
        font=(widgetfont, 12), padx=30, pady=10, bd=5)
        self.run_button.grid(column=1, row=16)

        self.save_button = tk.Button(self.root,
        text='Save', command=self.save_script,
        font=(widgetfont, 12), padx=30, pady=10, bd=5)
        self.save_button.grid(column=2, row=16)

        self.root.mainloop()

    # --- read enery scan matrix file
    def browse_file(self):
        self.escan_file = fd.askopenfilename(parent=self.root, 
        title='Choose energy scan matrix file (row: energy, column: time)',
        filetypes=(('any files', '*'), ('text files', '*.txt'), ('data files', '*.dat')))
        self.eps_file = fd.askopenfilename(parent=self.root, 
        title='Choose error matrix file of energy scan (row: energy, column: time)',
        filetypes=(('any files', '*'), ('text files', '*.txt'), ('data files', '*.dat')))
        self.time_file = fd.askopenfilename(parent=self.root, 
        title='Choose file for time delay array of energy scan matrix',
        filetypes=(('any files', '*'), ('text files', '*.txt'), ('data files', '*.dat')))
        self.print_file_num.delete('all')
        self.print_file_num.create_text(200, 15, text='files are successfully loaded')

        self.fit = False
        tmp = np.genfromtxt(self.escan_file)
        self.t = np.genfromtxt(self.time_file)
        self.escan_mat = tmp[:, 1:]
        self.e = tmp[:, 0]
        self.eps = np.genfromtxt(self.eps_file)

        self.U, self.sigma, self.Vh = svd(self.escan_mat, full_matrices=False)

    def plot_file(self):

        if self.escan_file is None or self.time_file is None:
            msg.showerror('Error', 'Please load escan matrix file and time delay array file')
        else:
            PlotDataWidget(self)
    
    def handle_float(self, entry, entry_name):
        if entry.get():
            test = entry.get()
            if isfloat.match(test):
                value = float(test)
            else:
                msg.showerror('Error',
                f'{entry_name} should be single float number')
                return False
        else:
            msg.showerror('Error',
            f'Please enter {entry_name}')
            return False
        return value

    def handle_float_field(self, entry, entry_name):
        if entry.get():
            tst = entry.get()
            if float_sep_comma.match(tst):
                value = np.array(list(map(float, tst.split(','))))
            else:
                msg.showerror('Error',
                f'{entry_name} should be single float or floats seperated by comma.')
                return False
        else:
            msg.showerror('Error',
            f'Please enter {entry_name}')
            return False
        return value
                
    def handle_irf(self):
        eta = None
        if self.irf_var.get() == 'g':
            fwhm = self.handle_float(self.entry_fwhm_G, 'fwhm_G')
        elif self.irf_var.get() == 'c':
            fwhm = self.handle_float(self.entry_fwhm_L, 'fwhm_L')
        elif self.irf_var.get() == 'pv':
            if self.entry_fwhm_G.get() and self.entry_fwhm_L.get():
                fwhm_G_init = self.entry_fwhm_G.get()
                fwhm_L_init = self.entry_fwhm_L.get()
                if isfloat.match(fwhm_G_init) and isfloat.match(fwhm_L_init):
                    fwhm = calc_fwhm(float(fwhm_G_init), float(fwhm_L_init))
                    eta = calc_eta(float(fwhm_G_init, fwhm_L_init))
                else:
                    msg.showerror('Error',
                    'Both fwhm_G and fwhm_L field should be single float number')
                    return False
            else:
                msg.showerror('Error',
                'Please enter both fwhm_G and fwhm_L values')
                return False
        else:
            msg.showerror('Error',
            'Please select irf model')
            return
        return fwhm, eta

    def handle_t0(self):
        t0 = self.handle_float(self.entry_t0, 't0')
        return t0

    def handle_tau(self):
        if self.entry_tau.get():
            str_tau = self.entry_tau.get()
            if float_sep_comma.match(str_tau):
                tau = np.array(list(map(float, str_tau.split(','))))
            else:
                msg.showerror('Error',
                'life time constant tau should be single float or floats seperated by comma.')
                return False
        else:
            if self.base_var.get():
                return True
            else:
                msg.showerror('Error',
                'Please enter life time constant')
                return False
        return tau
    
    def svd_script(self):
        if self.escan_file and self.time_file:
            msg.showerror('Error', 
            'Please load escan matrix file and time delay array file')
        else:
            PlotSVDWidget(self)


    def run_script(self):

        # check files are loaded
        if self.escan_file and self.time_file:
            msg.showerror('Error', 
            'Please load escan matrix file and time delay array file')
            return
    

        # set fwhm
        irf = self.irf_var.get()
        fwhm, eta = self.handle_irf()
        if not fwhm:
            return

        # set t0 of energy scan
        t0 = self.handle_t0()
        if not t0:
            return
        

        base = self.base_var.get()
        self.tau = self.handle_tau()
        if isinstance(self.tau, np.ndarray) or self.tau:
            if not isinstance(self.tau, np.ndarray):
                self.tau = None
        else:
            return
        
        cond_num = self.handle_float(self.entry_cond_num, 'Conditional Number')
        
        # reconstruct escan matrix based on SVD turncation filter
        N_servived = np.sum((self.sigma>cond_num*self.sigma[0]))
        escan_mat_turn = np.einsum('j,ij->ij', 
        self.sigma[:N_servived], self.U[:, :N_servived]) @ self.Vh[:N_servived, :]
        
        dads_spec, _, self.escan_fit = dads(self.t-t0, fwhm, self.tau, base, 
        irf=irf, eta=eta, intensity=escan_mat_turn)
        self.dads = dads_spec.T
        self.fit = True
        PlotDADSWidget(self)
        return
    
    def save_script(self):

        if self.escan_file and self.time_file:
            msg.showerror('Error', 'Please load escan matrix file and time delay array file')
            return
        
        if not self.fit:
            msg.showerror('Error', 'Please click run before saving results')
            return 

        save_directory = fd.askdirectory(parent=self.root, title='Choose directory to save results')
        np.savetxt(f'{save_directory}/LSVD.txt', 
        np.vstack((self.e, self.U.T)).T)
        np.savetxt(f'{save_directory}/RSVD.txt', 
        np.vstack((self.t, self.Vh)).T)
        np.savetxt(f'{save_directory}/sigma.txt', 
        self.sigma)
        np.savetxt(f'{save_directory}/dads.txt', 
        np.vstack((self.e, self.dads.T)).T)
        np.savetxt(f'{save_directory}/dads_fit.txt', 
        np.vstack((self.e, self.escan_fit.T)).T)
        return

def calc_dads_gui():
    CalcDADSGuiWidget()

if __name__ == '__main__':
    calc_dads_gui()





'''
_ampgo:
Adaptive Memory Programing For Global Optimization
Based On Andrea Gavana's Implementation (see: http://infinity77.net/global_optimization/)

For implementation detail see the `Tabu Tunneling Method` section in the below paper.
L. Lasdon et al. Computers & Operations Research 37 (2010) 1500–1509
http://leeds-faculty.colorado.edu/glover/fred%20pubs/416%20-%20AMP%20(TS)%20for%20Constrained%20Global%20Opt%20w%20Lasdon%20et%20al%20.pdf

:copyright: 2021-2022 by pistack (Junho Lee).
'''

from typing import Callable, Optional, Union
import numpy as np
from scipy.optimize import minimize, OptimizeResult

SCIPY_LOCAL_SOLVER = ['Nelder-Mead', 'Powell', 'CG', 'BFGS', 'Newton-CG',
'L-BFGS-B', 'TNC', 'COBYLA', 'SLSQP', 'dogleg', 'trust-ncg']


def ampgo(fun: Callable, x0: np.ndarray,
          tot_iter: Optional[int] = 20, max_tunnel: Optional[int] = 5,
          tol_tunnel: Optional[float] = 1e-5,
          minimizer_kwargs: Optional[dict] = None,
          eps1: Optional[float] = 0.02, eps2: Optional[float] = 0.1,
          n_tabu: Optional[int] = 5, strategy: Optional[str] = 'farthest',
          seed: Optional[Union[int, np.random.RandomState]] = None,
          callback: Optional[Callable] = None,
          disp: Optional[bool] = False) -> OptimizeResult:
    '''
    ampgo: Adaptive Memory Programming for Global Optimization
     Based on Tabu Tunneling Method

    Args:
     fun: Objective function.
      Objective function should have following form `f(x, *args)`
     x0: initial guess
     tot_iter: maximum number of global iteration
     max_tunnel: maximum number of tunneling phase
     tol_tunnel: Tolerance to determine whether tunneling phase is
      successful or not.
      If :math:`f(x_{best}) \\geq 0` and :math:`f(x_{tunnel}) < (1+{tol})f(x_{best})+{tol}` or
      :math:`f(x_{best}) < 0` and :math:`f(x_{tunnel}) < (1-{tol})f(x_{best})+{tol}` then
      such tunneling phase is regarded as successful phase.
     minimizer_kwargs: Extra keyword arguments to be passed to the local minimizer
      `scipy.optimize.minimize`. Some important options could be:

        * method (str): The minimization Method (default: `L-BFGS-B`)
        * args (tuple): The extra arguments passed to the objective function (`fun`) and its derivatives (`jac`, `hess`)
        * jac: jacobian of objective function (see scipy.optimize.minimize)
        * hess: hessian of objective function (see scipy.optimize.minimize)
        * bounds (Sequence of Tuple): Boundary of variable (see scipy.optimize.minimize)

     eps1: Constant used to define aspiration value
     eps2: Perturbation factor used to move away from the latest local minimum
     n_tabu: size of tabulist
     strategy ({'farthest', 'oldest'}): The strategy to delete element of tabulist when
      the size of tabulist exceeds `n_tabu`.

      * `farthest`: Delete farthest point from the latest local minimum point
      * `oldest`: Delete oldest point
     seed: seed used for random perturbation of local minimum. This argument is useful when
      someone wants to reproduce optimization results.
     callback: callback function to monitor each global iteration and tunneling phase.
      function signature should be `callback(x, f_val)`
     disp: display level, If zero or `None`, then no output is printed on screen.
      If postive number then status messages are printed.


    Returns:
     The optimization results represented as a `scipy.OptimizeResult` object.
     The important attributes are

     * `x`: The solution of the optimization
     * `fun`: values of objective fuction.
     * `success`: Whether or not the optimizer exited successfuly
     * `message`: Description of the cause of the termination

    Note:
     The implementation of ampgo method is based on
     L. Lasdon et al. Computers & Operations Research 37 (2010) 1500–1509. and
     Andrea Gavana's Python Implementation.
    '''

    # initialize
    nfev = 0
    total_tunnel = 0
    success_tunnel = 0
    f_best = np.inf
    tabulist = []
    tabu_size = 0
    x0 = np.atleast_1d(x0)
    n_param = x0.size
    ub = np.empty(n_param)
    lb = np.empty(n_param)

    monitor = False

    if isinstance(seed, np.random.RandomState):
        rng = seed
    elif isinstance(seed, int):
        rng = np.random.RandomState(seed)
    elif seed is None:
        rng = np.random.RandomState(None)
    else:
        raise Exception('Invalid seed for random generator')

    if callable(callback):
        monitor = True

    # Setting local minimizer
    if minimizer_kwargs is None:
        method = 'L-BFGS-B'
        args = ()
        bounds = None
        jac = None
        hess = None
        hessp = None
        tol = 1e-8
        minimizer_kwargs = {}
    else:
        method = minimizer_kwargs.pop('method', 'L-BFGS-B')
        args = minimizer_kwargs.pop('args', ())
        bounds = minimizer_kwargs.pop('bounds', None)
        jac = minimizer_kwargs.pop('jac', None)
        hess = minimizer_kwargs.pop('hess', None)
        hessp = minimizer_kwargs.pop('hessp', None)
        tol = minimizer_kwargs.pop('tol', 1e-8)

    if method not in SCIPY_LOCAL_SOLVER:
        raise Exception('Invalid local solver, local solver should be one of' +
        '[' + ', '.join(SCIPY_LOCAL_SOLVER) + ']')

    # Setting Arguments For tabu tunneling Function
    if callable(jac):
        ttf = wrapper_tunnel(fun, *args)
        jac_ttf = wrapper_grad_tunnel(fun, jac, *args)
    elif jac is True:
        ttf = wrapper_fun_grad_tunnel(fun, *args)
        jac_ttf = True
    else:
        ttf = wrapper_tunnel(fun, *args)
        jac_ttf = None

    if bounds is None:
        for i in range(n_param):
            ub[i] = np.inf
            lb[i] = -np.inf
    else:
        if len(bounds) != n_param:
            raise Exception(
                'Length of Bounds and the number of paramter should be same')
        for i in range(n_param):
            if bounds[i] is None:
                ub[i] = np.inf
                lb[i] = -np.inf
            else:
                lb[i], ub[i] = bounds[i]
                if lb[i] is None:
                    lb[i] = -np.inf
                if ub[i] is None:
                    ub[i] = np.inf

    # Start main loop
    for i in range(tot_iter+1):
        res = minimize(fun, x0, args=args, method=method,
                       jac=jac, hess=hess, hessp=hessp, bounds=bounds, tol=tol,
                       **minimizer_kwargs)
        f_opt = res['fun']
        x0 = res['x']
        nfev = nfev + res['nfev']
        if monitor:
            callable(x0, f_opt)
        if disp:
            print('='*72)
            print(f'local minimum is found in global iteration: {i}')
            if not res['success']:
                print('Warning: local optimization is failed')
        if f_opt < f_best:
            if disp:
                print(
                    f'local minimum improves solution at global iteration: {i}')
                print(f'Current: {f_opt} | previous: {f_best}')
            f_best = f_opt
            x_best = x0

        # If size of tabulist exceeds n_tabu then delete element
        # Add local minimum solution to the tabulist

        if tabu_size > n_tabu-1:
            tabulist = delete_element(x0, tabulist, strategy)
            tabu_size = n_tabu-1
        tabulist.append(x0)
        tabu_size = tabu_size+1

        # Calculates Aspiration
        aspiration = f_best - eps1*(1+np.abs(f_best))

        success = False
        for j in range(max_tunnel):
            total_tunnel = total_tunnel+1
            if disp:
                print('='*72)
                print(f'Tunneling Phase is Started: {i}-{j+1}')

            # Perturbe local minimum point x0
            vaild = False
            while not vaild:
                r = rng.uniform(-1, 1, n_param)
                beta = (eps2*(1+np.linalg.norm(x0)))/np.linalg.norm(r)
                x_try = x0 + beta*r
                x_try = np.where(x_try < lb, lb, x_try)
                x_try = np.where(x_try > ub, ub, x_try)
                vaild = check_vaild(x_try, tabulist)

            # start tabu tunneling
            res = minimize(ttf, x_try, args=(aspiration, tabulist), method=method,
                           jac=jac_ttf, bounds=bounds, tol=tol, **minimizer_kwargs)
            x0 = res['x']
            nfev = nfev + res['nfev']
            if jac is True:
                f_opt, _ = fun(x0, *args)
            else:
                f_opt = fun(x0, *args)
            nfev = nfev+1

            if f_best >= 0 and f_opt < (1+tol_tunnel)*f_best+tol_tunnel:
                success = True
                success_tunnel = success_tunnel+1
            elif f_best < 0 and f_opt < (1-tol_tunnel)*f_best+tol_tunnel:
                success = True
                success_tunnel = success_tunnel+1

            if monitor:
                callback(x0, f_opt)

            if disp and not res['success']:
                print('Warning: local optimization is failed')

            if disp and success:
                print('Tunneling phase is successful')

            if f_opt < f_best:
                if disp:
                    print(f'Tunneling phase {i}-{j+1} improves solution')
                    print(f'Current: {f_opt} | previous: {f_best}')
                f_best = f_opt
                x_best = x0

            # update tabulist
            if tabu_size > n_tabu-1:
                delete_element(x0, tabulist, strategy)
                tabu_size = n_tabu-1
            tabulist.append(x0)
            tabu_size = tabu_size+1

            if success:
                break

    # Optimization Process is finished

    result = OptimizeResult()
    result['fun'] = f_best
    result['x'] = x_best
    result['nfev'] = nfev
    result['success'] = True
    result['message'] = \
        ['Requested Number of global iteration is finished.']

    return result


def check_vaild(x_try, tabulist):
    '''
    Check random pertubation of latest local mimum point
    is vaild.
    '''
    dist = np.sum((tabulist-x_try)**2, axis=1)
    min_dist = np.min(dist)
    return min_dist > 1e-16


def delete_element(x_local, tabulist, strategy):
    '''
    Delete element from tabulist
    '''
    if strategy == 'oldest':
        tabulist.pop(0)
    else:
        dist = np.sum((tabulist-x_local)**2, axis=1)
        idx = np.argmax(dist)
        tabulist.pop(idx)

    return tabulist

def wrapper_tunnel(fun, *fun_args):
    '''
    wrapper function for tabu tunneling function
    '''
    def tunnel(x0, aspiration, tabulist):
        numerator = (fun(x0, *fun_args)-aspiration)**2
        denumerator = 1
        for tabu in tabulist:
            denumerator = denumerator*np.sum((x0-tabu)**2)
        return numerator/denumerator
    return tunnel

def wrapper_grad_tunnel(fun, jac, *fun_args):
    '''
    wrapper function for gradient of tabu tunneling function
    '''
    def grad_tunnel(x0, aspiration, tabulist):
            fval = fun(x0, *fun_args) - aspiration
            numerator = fval**2
            grad_numerator = fval*jac(x0, *fun_args)
            denominator = 1
            grad_denom = np.zeros_like(x0)
            for tabu in tabulist:
                diff = tabu-x0
                dist = np.sum(diff**2)
                denominator = denominator*dist
                grad_denom = grad_denom + diff/dist
            return 2*(grad_numerator+numerator*grad_denom)/denominator
    return grad_tunnel

def wrapper_fun_grad_tunnel(fun, *fun_args):
    '''
    wrapper function for pair of tabu tunneling function and
    its gradient
    '''
    def fun_grad_tunnel(x0, aspiration, tabulist):
        f_val, grad_val = fun(x0, *fun_args)
        f_val = f_val-aspiration
        numerator = f_val**2
        grad_numerator = f_val*grad_val
        denominator = 1
        grad_denominator = np.zeros_like(x0)
        for tabu in tabulist:
            diff = tabu-x0
            dist = np.sum(diff**2)
            denominator = denominator*dist
            grad_denominator = grad_denominator + \
                diff/dist
        
        y_ttf = numerator/denominator
        deriv_y_ttf = 2*(grad_numerator/denominator +
        y_ttf*grad_denominator)
        return y_ttf, deriv_y_ttf
    
    return fun_grad_tunnel



'''
_transient_exp:
submodule for fitting time delay scan with the
convolution of sum of exponential decay and instrumental response function

:copyright: 2021-2022 by pistack (Junho Lee).
:license: LGPL3.
'''
from typing import Optional, Union, Sequence, Tuple
import numpy as np
from ..mathfun.irf import calc_eta, calc_fwhm
from .transient_result import TransientResult
from ._ampgo import ampgo
from scipy.optimize import basinhopping
from scipy.optimize import least_squares
from ..mathfun.A_matrix import make_A_matrix_exp, fact_anal_A
from ..res.parm_bound import set_bound_t0, set_bound_tau
from ..res.res_decay import residual_decay, res_grad_decay
from ..res.res_decay import residual_decay_same_t0, res_grad_decay_same_t0

GLBSOLVER = {'basinhopping': basinhopping, 'ampgo': ampgo}


def fit_transient_exp(irf: str, fwhm_init: Union[float, np.ndarray],
                      t0_init: np.ndarray, tau_init: np.ndarray, base: bool,
                      method_glb: Optional[str] = None,
                      method_lsq: Optional[str] = 'trf',
                      kwargs_glb: Optional[dict] = None,
                      kwargs_lsq: Optional[dict] = None,
                      bound_fwhm: Optional[Sequence[Tuple[float, float]]] = None,
                      bound_t0: Optional[Sequence[Tuple[float, float]]] = None,
                      bound_tau: Optional[Sequence[Tuple[float, float]]] = None,
                      same_t0: Optional[bool] = False,
                      name_of_dset: Optional[Sequence[str]] = None,
                      t: Optional[Sequence[np.ndarray]] = None,
                      intensity: Optional[Sequence[np.ndarray]] = None,
                      eps: Optional[Sequence[np.ndarray]] = None) -> TransientResult:
    '''
    driver routine for fitting multiple data set of time delay scan data with
    sum of the convolution of exponential decay and instrumental response function.

    It separates linear and non-linear part of the optimization problem to solve non linear least sequare
    optimization problem efficiently.

    Moreover this driver uses two step method to search best parameter, its covariance and
    estimated parameter error.

    Step 1. (basinhopping)
    Use global optimization to find rough global minimum of our objective function.
    In this stage, it use analytic gradient for scalar residual function.

    Step 2. (method_lsq)
    Use least squares optimization algorithm to refine global minimum of objective function and approximate covariance matrix.
    Because of linear and non-linear seperation scheme, the analytic jacobian for vector residual function is hard to obtain.
    Thus, in this stage, it uses numerical jacobian.

    Args:
     irf ({'g', 'c', 'pv'}): shape of instrumental response function

      'g': gaussian shape

      'c': cauchy shape

      'pv': pseudo voigt shape (kind 2)
     fwhm_init (float or np.ndarray): initial full width at half maximum for instrumental response function

      * if irf in ['g', 'c'] then fwhm_init is float
      * if irf == 'pv' then fwhm_init is the `numpy.ndarray` with [fwhm_G, fwhm_L]

     t0_init (np.ndarray): time zeros for each scan
     tau_init (np.ndarray): lifetime constant for each decay component
     base (bool): Whether or not include baseline feature (i.e. very long lifetime constant)
     method_glb ({None, 'basinhopping', 'ampgo'}): Method for global optimization
     method_lsq ({'trf', 'dogbox', 'lm'}): method of local optimization for least_squares
                                           minimization (refinement of global optimization solution)
     kwargs_glb: keyward arguments for global optimization solver
     kwargs_lsq: keyward arguments for least square optimization solver
     bound_fwhm (sequence of tuple): boundary for irf parameter. If upper and lower bound are same,
      driver assumes that the parameter is fixed during optimization. If `bound_fwhm` is `None`,
      the upper and lower bound are given as `(fwhm_init/2, 2*fwhm_init)`.
     bound_t0 (sequence of tuple): boundary for time zero parameter.
      If `bound_t0` is `None`, the upper and lower bound are given by ``set_bound_t0``.
     bound_tau (sequence of tuple): boundary for lifetime constant for decay component,
      if `bound_tau` is `None`, the upper and lower bound are given by ``set_bound_tau``.
     same_t0 (bool): Whether or not time zero of every time delay scan in the same dataset should be same
     name_of_dset (sequence of str): name of each dataset
     t (sequence of np.narray): time scan range for each datasets
     intensity (sequence of np.ndarray): sequence of intensity of datasets for time delay scan
     eps (sequence of np.ndarray): sequence of estimated errors of each dataset

     Returns:
      TransientResult class object
    '''

    if method_glb is not None and method_glb not in ['basinhopping', 'ampgo']:
        raise Exception('Unsupported global optimization Method, Supported global optimization Methods are ampgo and basinhopping')
    if method_lsq not in ['trf', 'lm', 'dogbox']:
        raise Exception('Invalid local least square minimizer solver. It should be one of [trf, lm, dogbox]')
    if irf is not None and irf not in  ['g', 'c', 'pv']:
        raise Exception('Unsupported shape of instrumental response function Edge.')

    if tau_init is None:
        num_comp = 0
    else:
        num_comp = tau_init.size

    num_irf = 1*(irf in ['g', 'c'])+2*(irf == 'pv')
    num_param = num_irf+t0_init.size+num_comp
    param = np.empty(num_param, dtype=float)
    fix_param_idx = np.empty(num_param, dtype=bool)

    param[:num_irf] = fwhm_init
    param[num_irf:num_irf+t0_init.size] = t0_init
    param[num_irf+t0_init.size:] = tau_init
    bound = num_param*[None]

    if bound_fwhm is None:
        for i in range(num_irf):
            bound[i] = (param[i]/2, 2*param[i])
    else:
        bound[:num_irf] = bound_fwhm

    if bound_t0 is None:
        for i in range(t0_init.size):
            bound[i+num_irf] = set_bound_t0(t0_init[i], fwhm_init)
    else:
        bound[num_irf:num_irf+t0_init.size] = bound_t0

    if bound_tau is None:
        for i in range(num_comp):
            bound[i+num_irf +
                  t0_init.size] = set_bound_tau(tau_init[i], fwhm_init)
    else:
        bound[num_irf+t0_init.size:] = bound_tau

    for i in range(num_param):
        fix_param_idx[i] = (bound[i][0] == bound[i][1])

    if method_glb is not None:
        go_args = (num_comp, base, irf, fix_param_idx,
                   t, intensity, eps)
        min_go_kwargs = {'args': go_args, 'jac': True, 'bounds': bound}
        if kwargs_glb is not None:
            minimizer_kwargs = kwargs_glb.pop('minimizer_kwargs', None)
            if minimizer_kwargs is None:
                kwargs_glb['minimizer_kwargs'] = min_go_kwargs
            else:
                minimizer_kwargs['args'] = min_go_kwargs['args']
                minimizer_kwargs['jac'] = min_go_kwargs['jac']
                minimizer_kwargs['bounds'] = min_go_kwargs['bounds']
                kwargs_glb['minimizer_kwargs'] = minimizer_kwargs
        else:
            kwargs_glb = {'minimizer_kwargs': min_go_kwargs}
        if same_t0:
            res_go = GLBSOLVER[method_glb](res_grad_decay_same_t0, param, **kwargs_glb)
        else:
            res_go = GLBSOLVER[method_glb](res_grad_decay, param, **kwargs_glb)
    else:
        res_go = {}
        res_go['x'] = param
        res_go['message'] = None
        res_go['nfev'] = 0

    param_gopt = res_go['x']
    args_lsq = (base, irf, t, intensity, eps)

    if kwargs_lsq is not None:
        _ = kwargs_lsq.pop('args', None)
        _ = kwargs_lsq.pop('kwargs', None)
        kwargs_lsq['args'] = args_lsq
    else:
        kwargs_lsq = {'args': args_lsq}

    bound_tuple = (num_param*[None], num_param*[None])
    for i in range(num_param):
        bound_tuple[0][i] = bound[i][0]
        bound_tuple[1][i] = bound[i][1]
        if bound[i][0] == bound[i][1]:
            if bound[i][0] > 0:
                bound_tuple[1][i] = bound[i][1]*(1+1e-8)+1e-16
            else:
                bound_tuple[1][i] = bound[i][1]*(1-1e-8)+1e-16

    if same_t0:
        res_lsq = least_squares(residual_decay_same_t0, param_gopt,
        method=method_lsq, bounds=bound_tuple, **kwargs_lsq)
    else:
        res_lsq = least_squares(residual_decay, param_gopt,
        method=method_lsq, bounds=bound_tuple, **kwargs_lsq)
    param_opt = res_lsq['x']

    fwhm_opt = param_opt[:num_irf]
    tau_opt = param_opt[num_irf+t0_init.size:]

    fit = np.empty(len(t), dtype=object)
    res = np.empty(len(t), dtype=object)

    num_tot_scan = 0
    for i in range(len(t)):
        num_tot_scan = num_tot_scan + intensity[i].shape[1]
        fit[i] = np.empty(intensity[i].shape)
        res[i] = np.empty(intensity[i].shape)


# Calc individual chi2
    chi = res_lsq['fun']
    num_param_tot = num_tot_scan * \
        (num_comp+1*base)+num_param-np.sum(fix_param_idx)
    chi2 = 2*res_lsq['cost']
    red_chi2 = chi2/(chi.size-num_param_tot)

    start = 0
    end = 0
    chi2_ind = np.empty(len(t), dtype=object)
    red_chi2_ind = np.empty(len(t), dtype=object)
    num_param_ind = 2*tau_opt.size+1*base+2+1*(irf == 'pv')

    for i in range(len(t)):
        step = intensity[i].shape[0]
        chi2_ind_aux = np.empty(intensity[i].shape[1], dtype=float)
        for j in range(intensity[i].shape[1]):
            end = start + step
            chi2_ind_aux[j] = np.sum(chi[start:end]**2)
            start = end
        chi2_ind[i] = chi2_ind_aux
        red_chi2_ind[i] = chi2_ind[i]/(intensity[i].shape[0]-num_param_ind)

    param_name = np.empty(param_opt.size, dtype=object)
    c = np.empty(len(t), dtype=object)
    t0_idx = num_irf

    if irf == 'g':
        eta = 0
        fwhm_pv = fwhm_opt[0]
        param_name[0] = 'fwhm_G'
    elif irf == 'c':
        eta = 1
        fwhm_pv = fwhm_opt[0]
        param_name[0] = 'fwhm_L'
    else:
        eta = calc_eta(fwhm_opt[0], fwhm_opt[1])
        fwhm_pv = calc_fwhm(fwhm_opt[0], fwhm_opt[1])
        param_name[0] = 'fwhm_G'
        param_name[1] = 'fwhm_L'

    for i in range(len(t)):
        c[i] = np.empty((num_comp+1*base, intensity[i].shape[1]))

        if same_t0:
            A = \
                 make_A_matrix_exp(t[i]-param_opt[t0_idx],
                 fwhm_pv, tau_opt, base, irf, eta)


        for j in range(intensity[i].shape[1]):
            if not same_t0:
                A = \
                    make_A_matrix_exp(t[i]-param_opt[t0_idx],
                    fwhm_pv, tau_opt, base, irf, eta)
            c[i][:, j] = fact_anal_A(A, intensity[i][:, j], eps[i][:, j])
            fit[i][:, j] = c[i][:, j] @ A
            if not same_t0:
                param_name[t0_idx] = f't_0_{i+1}_{j+1}'
                t0_idx = t0_idx + 1

        if same_t0:
            param_name[t0_idx] = f't_0_{i}'
            t0_idx = t0_idx + 1

        res[i] = intensity[i] - fit[i]

    for i in range(num_comp):
        param_name[num_irf+t0_init.size+i] = f'tau_{i+1}'

    jac = res_lsq['jac']
    hes = jac.T @ jac
    cov = np.zeros_like(hes)
    n_free_param = np.sum(~fix_param_idx)
    mask_2d = np.einsum('i,j->ij', ~fix_param_idx, ~fix_param_idx)
    cov[mask_2d] = np.linalg.inv(hes[mask_2d].reshape(
        (n_free_param, n_free_param))).flatten()
    cov_scaled = red_chi2*cov
    param_eps = np.sqrt(np.diag(cov_scaled))
    corr = cov_scaled.copy()
    weight = np.einsum('i,j->ij', param_eps, param_eps)
    corr[mask_2d] = corr[mask_2d]/weight[mask_2d]

    result = TransientResult()
    result['model'] = 'decay'
    result['same_t0'] = same_t0
    result['fit'] = fit
    result['res'] = res
    result['irf'] = irf
    result['eta'] = eta
    result['fwhm'] = fwhm_pv

    # save experimental fitting data
    if name_of_dset is None:
        name_of_dset = np.empty(len(t), dtype=object)
        for i in range(len(t)):
            name_of_dset[i] = f'dataset_{i+1}'
    result['name_of_dset'] = name_of_dset
    result['t'] = t
    result['intensity'] = intensity
    result['eps'] = eps

    result['param_name'] = param_name
    result['x'] = param_opt
    result['bounds'] = bound
    result['base'] = base
    result['c'] = c
    result['chi2'] = chi2
    result['chi2_ind'] = chi2_ind
    result['aic'] = chi.size*np.log(chi2/chi.size)+2*num_param_tot
    result['bic'] = chi.size * \
        np.log(chi2/chi.size)+num_param_tot*np.log(chi.size)
    result['red_chi2'] = red_chi2
    result['red_chi2_ind'] = red_chi2_ind
    result['nfev'] = res_go['nfev'] + res_lsq['nfev']
    result['n_param'] = num_param_tot
    result['n_param_ind'] = num_param_ind
    result['num_pts'] = chi.size
    result['jac'] = jac
    result['cov'] = cov
    result['corr'] = corr
    result['cov_scaled'] = cov_scaled
    result['x_eps'] = param_eps
    result['method_lsq'] = method_lsq
    result['message_lsq'] = res_lsq['message']
    result['success_lsq'] = res_lsq['success']

    if result['success_lsq']:
        result['status'] = 0
    else:
        result['status'] = -1

    if method_glb is not None:
        result['method_glb'] = method_glb
        result['message_glb'] = res_go['message'][0]
    else:
        result['method_glb'] = None
        result['message_glb'] = None

    result['n_osc'] = 0
    if tau_init is None:
        result['n_decay'] = 0
    else:
        result['n_decay'] = tau_init.size

    return result
